#! /usr/bin/env python
#
"""
Driver script for running 'Value Function Iteration' programs
on the SMSCG infrastructure.
"""
# Copyright (C) 2011-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
# summary of user-visible changes
__changelog__ = """
  2011-05-06:
    * Workaround for Issue 95: now we have complete interoperability
      with GC3Utils.
"""
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import george
    george.GeorgeScript().run()


# stdlib imports
import ConfigParser
import csv
import math
import os
import os.path
import shutil
import sys
from prettytable import PrettyTable
import types

# interface to Gc3libs
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, nonnegative_int, positive_int
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection
import gc3libs.utils

IN_VALUES_FILE = 'ValuesIn.txt'
OUT_VALUES_FILE = 'ValuesOut.txt'
TMPDIR = '/tmp'


## support classes

# For each input file, a new computation is started, totally
# `self.params.iterations` passes, each pass corresponding to
# an application of the `self.params.executable` function.
#
# This is the crucial point:

class ValueFunctionIteration(SequentialTaskCollection):
    """
    Perform a (predefined) number of iterations of a certain function
    over a set of values.  The function to be iterated is implemented
    in the form of an executable program, that takes a single input
    file IN_VALUES_FILE (a list of values) and creates a single output
    file OUT_VALUES_FILE.

    The several-passes computation is implemented as a sequential task
    collection (totalling `total_iterations` steps); each step of
    which is a parallel collection of tasks, each of which is a
    single-core task executing the given program.
    """

    def __init__(self, executable, initial_values_file,
                 total_iterations, slice_size=0,
                 output_dir=TMPDIR, **extra_args):
        """
        Create a new task that runs `executable` over a set of values
        (initially given by `initial_values_file`, then the output of
        a run is fed as input to the next one), iterating the process
        `total_iterations` times.

        If `slice_size` is a positive integer, then chop the input into
        chunks of -at most- the given size and compute them as separate
        independent jobs.

        Extra keyword arguments are saved and passed down to construct
        the `ValueFunctionIterationApplication`.
        """
        assert slice_size >= 0, \
               "Argument `slice_size` to ValueFunctionIteration.__init__" \
               " must be a non-negative integer."

        # remember values for later use
        self.executable = executable
        self.initial_values = initial_values_file
        self.total_iterations = total_iterations - 1
        self.slice_size = slice_size
        self.datadir = output_dir
        self.extra = extra_args

        # count initial values
        if os.path.exists(initial_values_file):
            f = open(initial_values_file, 'r')
            self.num_input_values = gc3libs.utils.count(f, lambda _: True)
            f.close()
        else:
            # XXX: there's no good default value!
            self.num_input_values = 0

        # this little piece of black magic is to ensure intermediate
        # filenames appear numerically sorted in `ls -l` output
        self.values_filename_fmt = ('values.%%0%dd.txt'
                                    % (1 + int(math.log10(total_iterations))))

        self.jobname = extra_args.get('jobname',
                              gc3libs.utils.basename_sans(initial_values_file))

        # create initial task and register it
        initial_task = ValueFunctionIterationPass(executable, initial_values_file,
                                                  0, total_iterations, slice_size,
                                                  self.datadir, self.extra,
                                                  parent=self.jobname)
        SequentialTaskCollection.__init__(self, self.jobname, [initial_task])


    def __str__(self):
        return self.jobname


    def next(self, iteration):
        """
        If there are more iterations to go, enqueue the corresponding jobs.

        See: `SequentialTaskCollection.next`:meth: for a description
        of the contract that this method must implement.
        """
        if iteration == self.total_iterations:
            return Run.State.TERMINATED
        else:
            last = self.tasks[iteration]
            ## write output values to a 'values.<iteration>.txt' file
            last_values_filename = os.path.join(self.datadir,
                                                self.values_filename_fmt % iteration)
            last_values_file = open(last_values_filename, 'w+')
            for value in last.output_values:
                last_values_file.write('%f\n' % value)
            last_values_file.close()
            ## now create new task for computing the next pass
            self.add(
                ValueFunctionIterationPass(
                    self.executable,
                    last_values_filename,
                    iteration+1,
                    self.total_iterations,
                    self.slice_size,
                    extra=self.extra,
                    parent=self.jobname,
                    )
                )
            return Run.State.RUNNING


    def terminated(self):
        """
        Collect the output of all iterations into a single '.csv'
        file.
        """
        # determine output file name
        output_filename_base = gc3libs.utils.basename_sans(self.initial_values)
        if output_filename_base.endswith('.input'):
            output_filename_base = output_filename_base[:-6]
        output_filename = output_filename_base  + '.output.csv'
        gc3libs.log.debug("SequentialTaskCollection %s done,"
                          " now processing results into file '%s'..."
                          % (self, output_filename))
        output_file = open(output_filename, 'w+') # write+truncate
        output_csv = csv.writer(output_file)
        output_csv.writerow(['i'] + [ ('n=%d' % n)
                                      for n in range(1, len(self.tasks)+1) ])
        for i, values in enumerate(zip(*[ task.output_values
                                          for task in self.tasks ])):
            output_csv.writerow([i] + list(values))
        output_file.close()
        # same stuff, albeit in a different output format for SimpleDP compatibility
        output_filename = output_filename_base + '.output.txt'
        shutil.copyfile(self.initial_values, output_filename)
        output_file = open(output_filename, 'a') # write, no truncate
        for task in self.tasks:
            for value in task.output_values:
                output_file.write("%s\n" % value)
        output_file.close()
        gc3libs.log.debug("  ...done.")


class ValueFunctionIterationPass(ParallelTaskCollection):
    """
    Compute the values taken by a certain function over a set of
    inputs.  The function to be iterated is implemented in the form of
    an executable program, that takes a single input file `IN_VALUES_FILE`
    (a list of values) and creates a single output file `OUT_VALUES_FILE`.

    The computation will be split into separate independent processes,
    each working over a fraction of the input values (determined by
    the `slice_size` argument to `__init__`).

    This is meant to be a single step in the value function iteration.
    """

    def __init__(self, executable, input_values_file,
                 iteration, total_iterations,
                 slice_size=0, datadir=TMPDIR, extra={ },
                 parent=None):
        """
        Create a new tasks that runs `executable` over the set of
        values contained in file `input_values_file` (one
        floating-point number per line).

        If `slice_size` is a positive integer, then chop the input into
        chunks of -at most- the given size and compute them as separate
        independent jobs.

        Any other argument is passed unchanged to the
        `ParallelTaskCollection` ctor.
        """
        assert slice_size >= 0, \
               "Argument `slice_size` to ValueFunctionIterationPass.__init__" \
               " must be a non-negative integer."
        assert isinstance(extra, dict), \
               "Argument `extra` to ValueFunctionIterationPass.__init__" \
               " must be a dictionary instance."

        self.input_values = input_values_file
        self.output_values = None

        total_input_values = _count_input_values(input_values_file)

        if slice_size < 1:
            # trick to make the for-loop below work in the case of one
            # slice only
            slice_size = total_input_values

        # pad numbers with correct amount of zeros, so they look
        # sorted in plain `ls -l` output
        fmt = '%%0%dd' % (1 + int(math.log10(float(total_iterations))))
        self.jobname = ("%s.%s"
                        % ((parent or gc3libs.utils.basename_sans(input_values_file)),
                           (fmt % iteration)))

        # create data sub-directory
        datasubdir = os.path.join(datadir, self.jobname)
        if not os.path.exists(datasubdir):
            os.makedirs(datasubdir)

        # build list of tasks
        tasks = [ ]
        for start in range(0, total_input_values, slice_size):
            # create new job to handle this slice of values
            extra_args = extra.copy()
            extra_args['parent'] = self.jobname
            tasks.append(
                ValueFunctionIterationApplication(
                    executable,
                    input_values_file,
                    iteration,
                    total_iterations,
                    # each task computes values with i in range
                    # `start..end` (inclusive), and `end` is
                    # generally `slice_size` elements after `start`
                    start,
                    end=min(start + slice_size - 1, total_input_values),
                    output_dir = datasubdir,
                    **extra_args
                    )
                )

        # actually init jobs
        ParallelTaskCollection.__init__(self, self.jobname, tasks)


    def __str__(self):
        return self.jobname


    def terminated(self):
        """
        Collect all results from sub-tasks into `self.output_values`.
        """
        gc3libs.log.debug(
            "%s terminated, now collecting return values from sub-tasks ..."
            % self)
        self.output_values = [ ]
        for task in self.tasks:
            if task.execution.exitcode != 0:
                self.execution.exitcode = 1
                gc3libs.log.error(
                    "%s: sub-task %s failed with exit code %d. Aborting."
                    % (self, task, task.execution.exitcode))
                return
            self.output_values.extend(task.output_values)
        gc3libs.log.debug("  ...done.")


class ValueFunctionIterationApplication(Application):
    """
    A computational job to reckon the values taken by a certain
    function over a set of inputs.  The function to be iterated is
    implemented in the form of an executable program, that takes a
    single input file IN_VALUES_FILE (a list of values) and creates a
    single output file OUT_VALUES_FILE.

    Optional arguments `start` and `end` constrain the application of
    the function to a subset of the input values (those with positions
    in the range `start..end`, inclusive of endpoints).

    When the job turns to TERMINATED state and the output is fetched
    back, the list of output values is made available in the
    `self.output_values` attribute.
    """

    application_name = 'vfi'

    def __init__(self, executable, input_values_file, iteration, total_iterations,
                 start=0, end=None, parent=None, **extra_args):
        count = _count_input_values(input_values_file)
        if end is None:
            end = count-1 # last allowed position
        # make a readable jobname, indicating what part of the computation this is
        extra_args.setdefault('jobname',
                      '%s.%d--%d' % (
                          parent or gc3libs.utils.basename_sans(input_values_file),
                          start, end))
        Application.__init__(
            self,
            os.path.basename(executable),
            # `start` and `end` arguments to `executable` follow the
            # FORTRAN convention of being 1-based
            arguments = [ extra_args['discount_factor'], iteration, total_iterations,
                          count, start+1, end+1 ],
            inputs = { executable:os.path.basename(executable),
                       input_values_file:IN_VALUES_FILE },
            outputs = { OUT_VALUES_FILE:OUT_VALUES_FILE },
            join = True,
            stdout = 'job.log', # stdout + stderr
            **extra_args)


    def __str__(self):
        return self.jobname


    def terminated(self):
        gc3libs.log.debug("%s: TERMINATED with return code %d ..." % (self.persistent_id, self.execution.returncode))
        if self.execution.returncode == 0:
            gc3libs.log.debug("%s: terminated correctly, now post-processing results ..." % self.persistent_id)
            # everything ok, try to post-process results
            results = [ ]
            output_dir = self.output_dir
            output_filename = os.path.join(output_dir, OUT_VALUES_FILE)
            try:
                result_file = open(output_filename, 'r')
                for lineno, line in enumerate(result_file):
                    # convert line to floating-point number
                    results.append(float(line))
                result_file.close()
                self.output_values = results
                # if we got to this point, parsing went fine
                # so we can remove the output directory altogether
                shutil.rmtree(output_dir, ignore_errors=True)
            except ValueError, ex:
                # some line cannot be parsed as a floating-point number
                msg = ("Invalid content in file '%s' at line %d: %s"
                       % (output_filename, lineno, str(ex)))
                gc3libs.log.error("%s: %s" % (self.persistent_id, msg))
                self.info = msg
                self.exitcode = 65 # EX_DATAERR in /usr/include/sysexits.h
            except IOError, ex:
                # error opening or reading file
                msg = ("I/O error processing output file '%s': %s"
                       % (output_filename, str(ex)))
                gc3libs.log.error("%s: %s" % (self.persistent_id, msg))
                self.info = msg
                self.exitcode = 74 # EX_IOERR in /usr/include/sysexits.h
            except Exception, ex:
                msg = ("Error processing result file '%s': %s"
                       % (output_filename, str(ex)))
                gc3libs.log.error("%s: %s" % (self.persistent_id, msg))
                self.info = msg
                self.exitcode = 70 # EX_SOFTWARE in /usr/include/sysexits.h


## auxiliary functions

def _count_input_values(input_values):
    """
    Return number of input values.  Argument `input_values` can be
    either a file name (in which case the count of lines in the file
    is returned, assuming that the file has one input value per line),
    or a Python sequence of values.
    """
    if isinstance(input_values, types.StringTypes):
        # `input_values` is the name of a file, count number of lines
        input_values_file = open(input_values, 'r')
        count = len(input_values_file.readlines())
        input_values_file.close()
        return count
    else:
        # assume `input_values` is a sequence of some type and return
        # its length
        return len(input_values)


## main

class GeorgeScript(SessionBasedScript):
    """
    Execute P iterations of the specified value function on the input
    data.  For each input file, all iterations are collected back into
    a single '.csv' output file: column K in the output file contains
    the results of the K-th iteration of the value function.

    Computation of the value function on a set of values is performed
    by a separate program ``vfi.exe``.  You can set an alternate path
    to the compute program using the ``-x`` command-line option.

    The number P of iterations can be set with the ``-P`` command-line
    option.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = '0.3',
            # only '.txt' files are considered as valid input
            input_filename_pattern = '*.ini',
            )

    def setup_args(self):
        super(GeorgeScript, self).setup_args()

        self.add_param('-P', '--iterations', metavar='NUM',
                       dest='iterations', type=positive_int, default=1,
                       help="Compute NUM iterations per each output file"
                       " (default: %(default)s).")
        self.add_param('-p', '--slice-size', metavar='NUM',
                       dest='slice_size', type=nonnegative_int, default=0,
                       help="Process at most NUM states in a single"
                       " computational job.  Each input file is chopped"
                       " into files that contain at most NUM elements,"
                       " each of which is processed separately as a single"
                       " computational jobs.  The higher NUM is, the less"
                       " jobs will be created, but the longer each job will"
                       " last; this value and the DURATION argument to"
                       " option `-w` are directly proportional."
                       " If 0 (default), no chopping takes place and one job"
                       " only is used for each pass of the computation.")
        self.add_param("-x", "--execute", dest="execute",
                       type=executable_file, default="vfi.exe",
                       metavar="EXECUTABLE",
                       help="Run the specified EXECUTABLE program"
                       " to compute one step of values;"
                       " default: %(default)s.")


    def parse_args(self):
        if not os.path.isabs(self.params.execute):
            self.params.execute = os.path.abspath(self.params.execute)


    def new_tasks(self, extra):
        inis = self._search_for_input_files(self.params.args)

        p = ConfigParser.SafeConfigParser()
        successfully_read = p.read(inis)
        for filename in inis:
            if filename not in successfully_read:
                self.log.error("Could not read/parse input file '%s'. Ignoring it."
                               % filename)

        for name in p.sections():
            # path to the initial values file
            if not p.has_option(name, 'initial_values_file'):
                self.log.error("Required parameter 'initial_values_file' missing"
                               " in input '%s'.  Ignoring."
                               % name)
            path = p.get(name, 'initial_values_file')
            if not os.path.isabs(path):
                path = os.path.abspath(path)
            if not os.path.exists(path):
                self.log.error("Input values file '%s' does not exist."
                               " Ignoring task '%s', which depends on it."
                               % (path, name))

            # import running parameters from cfg file
            extra_args = extra.copy()
            extra_args['jobname'] = name
            try:
                extra_args.setdefault('discount_factor',
                              p.getfloat(name, 'discount_factor'))
            except (KeyError, ConfigParser.Error, ValueError), ex:
                self.log.error("Could not read required parameter 'discount_factor'"
                               " in input '%s': %s"
                               % (name, str(ex)))

            yield (name, george.ValueFunctionIteration, [
                self.params.execute, path,
                self.params.iterations,
                self.params.slice_size
                ], extra_args)


    def print_summary_table(self, output, stats):
        table = PrettyTable(['Input', 'Iteration', 'Tasks Generated/Total', 'Progress'])
        table.align = 'c'
        table.align['Input'] = 'l'
        def compute_stats(collection):
            result = collection.stats()
            def add_stats(s1, s2):
                for k in s2.iterkeys():
                    s1[k] += s2[k]
                return s1
            for task in collection.tasks:
                if hasattr(task, 'stats'):
                    add_stats(result, task.stats())
            return result
        for toplevel in self.session:
            current_iteration = toplevel._current_task
            total_iterations = toplevel.total_iterations
            if toplevel.slice_size == 0:
                mult = 2
            else:
                mult = 1 + (toplevel.num_input_values / toplevel.slice_size)
            generated_tasks = 1 + mult * (1 + current_iteration)
            total_tasks = 1 + mult * (1 + total_iterations)
            progresses = [ ]
            for state in [ Run.State.NEW,
                           Run.State.SUBMITTED,
                           Run.State.RUNNING,
                           Run.State.STOPPED,
                           Run.State.TERMINATING,
                           Run.State.TERMINATED,
                           Run.State.UNKNOWN
                           ]:
                stats = compute_stats(toplevel)
                count = stats[state]
                if count > 0:
                    progresses.append("%.1f%% %s"
                                      % (100.0 * count / (total_tasks-1), state))
                    if state == 'TERMINATED':
                        progresses[-1] += (" (%.1f%% ok, %.1f%% failed)"
                                           % (100.0 * stats['ok'] / count,
                                              100.0 * stats['failed'] / count))
            table.add_row([
                toplevel.jobname,
                "%d/%d" % (1+current_iteration, 1+total_iterations),
                "%d/%d" % (generated_tasks, total_tasks),
                str.join(", ", progresses)
                ])
        output.write(str(table))
        output.write("\n")



#! /usr/bin/env python
#
#   gmhc_coev.py -- Front-end script for submitting multiple `MHC_coev` jobs to SMSCG.
#
#   Copyright (C) 2011-2012 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `MHC_coev` jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gmhc_coev --help`` for program usage instructions.
"""
__version__ = '2.1.0rc2 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2011-07-22:
    * Use the ``APPS/BIO/MHC_COEV-040711`` run time tag to select execution sites.
  2011-06-15:
    * Initial release, forked off the ``ggamess`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import gmhc_coev
    gmhc_coev.GMhcCoevScript().run()


# std module imports
import csv
import glob
import math
import os
import re
import shutil
import sys
import time

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from gc3libs.compat._collections import defaultdict
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import SequentialTaskCollection


## auxilirary functions

def _increase(value, inflection, ratio=2, increment=2):
    """
    Return increased `value`.

    The new value is `value` multiplied by `ratio` if the product is
    less than `inflection`; otherwise, it's `value` plus `increment`.
    In other words, values grow exponentially until the inflection
    point, after which they grow linearly.
    """
    if value > inflection:
        return (ratio * value)
    else:
        return (value + increment)


## custom application class

class GMhcCoevApplication(Application):
    """
    Custom class to wrap the execution of a single step of the
    ``MHC_coev_*`` program by A. B. Wilson and collaborators.
    """

    application_name = 'mhc_coev'

    def __init__(self,
                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                 output_dir, latest_work=None, executable=None, **extra_args):
        extra_args.setdefault('requested_memory', 1*GB)
        extra_args.setdefault('requested_cores', 1)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        # command-line parameters to pass to the MHC_coev_* program
        self.N = N
        self.p_mut_coeff = p_mut_coeff
        self.choose_or_rand = choose_or_rand
        self.sick_or_not = sick_or_not
        self.off_v_last = off_v_last
        if executable is not None:
            # use the specified executable
            executable_name = './' + os.path.basename(executable)
            inputs = { executable:os.path.basename(executable) }
        else:
            # use the default one provided by the RTE
            executable_name = '/$MHC_COEV'
            inputs = { }
        if latest_work is not None:
            inputs[latest_work] = 'latest_work.mat'
        Application.__init__(self, [
                                 executable_name,
                                 # use `single_run_time` as the
                                 # maximum allowed time before saving
                                 # the MatLab workspace, but allow 5
                                 # minutes for I/O before the job is
                                 # killed forcibly by the batch system
                                 extra_args.get('requested_walltime').amount(minutes) - 5,
                                 # rest of `MHC_coev` args as passed
                                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                             ],
                             inputs = inputs,
                             outputs = gc3libs.ANY_OUTPUT,
                             output_dir = output_dir,
                             stdout = 'matlab.log',
                             stderr = 'matlab.err',
                             tags = [ 'TEST/MHC_COEV-040711ML2012' ],
                             **extra_args)


class GMhcCoevTask(SequentialTaskCollection):
    """
    Custom class to wrap the execution of the ``MHC_coev_*` program
    by A. B. Wilson and collaborators.

    Execution continues in steps of predefined duration until a
    specified number of generations have been computed and saved.  All
    the output files ever produced are collected in the path specified
    by the `output_dir` parameter to `__init__`.

    """

    def __init__(self, single_run_duration, generations_to_do,
                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                 output_dir, executable=None, **extra_args):

        """Create a new task running an ``MHC_coev`` binary.

        Each binary is expected to run for `single_run_duration`
        minutes and dump its state in file ``latest_work.mat`` if it's
        not finished. This task will continue re-submitting the same
        executable together with the saved workspace until
        `generations_to_do` generations have been computed.

        :param single_run_duration:
          Duration of a single step (as a `gc3libs.quantity.Duration`:class: value)

        :param int generations_to_do:
          Count of generations that ``MHC_coev`` should simulate.

        :param              N: Passed unchanged to the MHC_coev program.
        :param    p_mut_coeff: Passed unchanged to the MHC_coev program.
        :param choose_or_rand: Passed unchanged to the MHC_coev program.
        :param    sick_or_not: Passed unchanged to the MHC_coev program.
        :param     off_v_last: Passed unchanged to the MHC_coev program.

        :param str output_dir:
          Path to a directory where output files
          from all runs should be collected.

        :param str executable:
          Path to the ``MHC_coev`` executable binary, or `None`
          (default) to specify that the default version available on
          the execution site should be used.

        """
        # remember values for later use
        self.executable = executable
        self.output_dir = output_dir
        # allow 5 extra minutes for final saving the MatLab workspace
        self.single_run_duration = single_run_duration + 5*minutes
        self.generations_to_do = generations_to_do
        self.p_mut_coeff = p_mut_coeff
        self.N = N
        self.choose_or_rand = choose_or_rand
        self.sick_or_not = sick_or_not
        self.off_v_last = off_v_last
        self.extra = extra_args

        self.generations_done = 0

        # make up a sensibel job name if there isn't one
        if 'jobname' in extra_args:
            self.jobname = extra_args['jobname']
        else:
            if self.executable is None:
                self.jobname = ('MHC_coev.%s.%s.%s.%s.%s'
                                % (N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last))
            else:
                os.path.basename(self.executable)

        # remove potentially conflicting kyword arguments
        extra_args.pop('output_dir', None)
        extra_args.pop('requested_walltime', None)

        # create initial task and register it
        initial_task = GMhcCoevApplication(
            N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
            output_dir = os.path.join(output_dir, 'tmp'),
            executable = self.executable,
            requested_walltime = self.single_run_duration,
            **extra_args)
        SequentialTaskCollection.__init__(self, [initial_task])


    # regular expression for extracting the generation no. from an output file name
    GENERATIONS_FILENAME_RE = re.compile(r'(?P<generation_no>[0-9]+)gen\.mat$')

    def next(self, done):
        """
        Analyze the retrieved output and decide whether to submit
        another run or not, depending on whether there is a
        ``latest_work.mat`` file.
        """
        task_output_dir = self.tasks[done].output_dir
        if not os.path.exists(task_output_dir):
            # no output, resubmit immediately
            return done
        stdout_filename = self.tasks[done].stdout
        stderr_filename = self.tasks[done].stderr
        last_run = self.tasks[done].execution
        exclude = [
            os.path.basename(self.tasks[done].arguments[0]),
            stdout_filename,
            stderr_filename,
            ]
        # move files one level up, except the ones listed in
        # `exclude`, and read the contents of STDOUT/STDERR into
        # `job_out` and `job_err`
        stdout_contents = ''
        stderr_contents = ''
        generation_files_count = 0
        for entry in os.listdir(task_output_dir):
            src_entry = os.path.join(task_output_dir, entry)
            # concatenate all output files together
            if entry == stdout_filename:
                gc3libs.utils.cat(src_entry, output=os.path.join(self.output_dir, entry), append=True)
                stdout_contents = gc3libs.utils.read_contents(src_entry)
            if entry == stderr_filename:
                gc3libs.utils.cat(src_entry, output=os.path.join(self.output_dir, entry), append=True)
                stderr_contents = gc3libs.utils.read_contents(src_entry)
            if entry in exclude or (entry.startswith('script.') and entry.endswith('.sh')):
                # delete entry and continue with next one
                os.remove(src_entry)
                continue
            # if `entry` is a generation output file, get the
            # generation no. and update the generation count
            match = GMhcCoevTask.GENERATIONS_FILENAME_RE.search(entry)
            if match:
                generation_files_count += 1
                generation_no = int(match.group('generation_no'))
                self.generations_done = max(self.generations_done, generation_no)
            # now really move file one level up
            dest_entry = os.path.join(self.output_dir, entry)
            if os.path.exists(dest_entry):
                # backup with numerical suffix
                gc3libs.utils.backup(dest_entry)
            os.rename(os.path.join(task_output_dir, entry), dest_entry)
        os.removedirs(task_output_dir)

        # scan for common failures and take correction measures
        report_filename = os.path.join(self.output_dir, stdout_filename)
        if os.path.exists(report_filename):
            report = open(report_filename, 'a')
        else:
            report = open(report_filename, 'w')
        report.write(
            "gmhc_coev: About last job: duration=%s, exitcode=%d, max_used_memory=%s\n"
            % (last_run.duration, last_run.exitcode, last_run.max_used_memory))


        # 0. did the job run at all?
        if last_run.duration < 5*minutes:
            report.write(
                "gmhc_coev: Job %s did not run at all,"
                " will try resubmitting to a different resource.\n"
                % (self,))
            # resubmit
            return done

        # 1. out of memory?
        elif (# generic memory error
            (last_run.max_used_memory >= self.extra['requested_memory'])
            # MATLAB-specific error condition
            or (last_run.exitcode == 255
                and ('Out of memory.' in stderr_contents
                     or 'MATLAB:nomem' in stderr_contents))):
            self.extra['requested_memory'] = _increase(self.extra['requested_memory'], 8*GB, increment=2*GB) # start growing linearly at 8GB
            report.write(
                "gmhc_coev: Possible out-of-memory condition detected in job %s,"
                " will request %s for next run.\n"
                % (self.tasks[done], self.requested_memory,))

        # 2. not enough time to compute even one generation?
        elif (generation_files_count == 0
              # be sure that the job has run for at least the expected time, +/- 30 minutes
              and (last_run.duration > (self.single_run_duration - 30*minutes))):
            self.single_run_duration = _increase(self.single_run_duration, 24*hours,
                                                 increase=4*hours)
            report.write(
                "gmhc_coev: Job %s could not complete 1 generation in the allotted time,"
                " will request %s for next run.\n"
                % (self.tasks[done], self.single_run_duration,))

        # if a `latest_work.mat` file exists, then we need
        # more time to compute the required number of generations
        latest_work = os.path.join(self.output_dir, 'latest_work.mat')
        if not os.path.exists(latest_work):
            latest_work = None
            report.write(
                "gmhc_coev: No `latest_work.mat` file was produced,"
                " next job will start from k=1.\n")
        if self.generations_done < self.generations_to_do:
            gc3libs.log.debug("Computed %s generations, %s to do; submitting another job."
                              % (self.generations_done, self.generations_to_do))
            report.write("gmhc_coev:"
                         " Computed %s generations, %s to do; submitting another job.\n"
                         % (self.generations_done, self.generations_to_do))
            report.close()
            self.add(
                GMhcCoevApplication(self.N, self.p_mut_coeff, self.choose_or_rand, self.sick_or_not, self.off_v_last,
                                    output_dir = os.path.join(self.output_dir, 'tmp'),
                                    latest_work = latest_work,
                                    executable = self.executable,
                                    requested_walltime = self.single_run_duration,
                                    **self.extra))
            return Run.State.RUNNING
        else:
            gc3libs.log.debug("Computed %s generations, no more to do."
                              % (self.generations_done))
            report.write("gmhc_coev:"
                         " Computed %s generations, no more to do.\n"
                         % (self.generations_done,))
            report.close()
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED


## main script class

class GMhcCoevScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for executable files
whose name starts with the string ``MHC_coev``, and submit a job for
each file found; job progress is monitored and, when a job is done,
its ``.sav`` output files are retrieved back into the same directory
where the executable file is (this can be overridden with the ``-o``
option).

The ``gmhc_coev`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gmhc_coev`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            input_filename_pattern = 'MHC_coev_*',
            application = GMhcCoevTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GMhcCoevTask,
            )

    def setup_options(self):
        self.add_param("-G", "--generations", metavar="NUM",
                       dest="generations", type=int, default=3000,
                       help="Compute NUM generations (default: 3000).")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=None,
                       help="Path to the MHC_coev_* executable file.")


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    def new_tasks(self, extra):
        # how many iterations are we already computing (per parameter set)?
        iters = defaultdict(lambda: 0)
        for task in self.session:
            name, instance = task.jobname.split('#')
            iters[name] = max(iters[name], int(instance))

        for path in self.params.args:
            if path.endswith('.csv'):
                try:
                    inputfile = open(path, 'r')
                except (OSError, IOError), ex:
                    self.log.warning("Cannot open input file '%s': %s: %s",
                                     path, ex.__class__.__name__, str(ex))
                try:
                    # the `csv.sniff()` function is confused by blank and comment lines,
                    # so we need to filter the input to build a correct sample
                    sample_lines = [ ]
                    while len(sample_lines) < 5:
                        line = inputfile.readline()
                        # exit at end of file
                        if line == '':
                            break
                        # ignore comment lines as they confuse `csv.sniff`
                        if line.startswith('#') or line.strip() == '':
                            continue
                        sample_lines.append(line)
                    csv_dialect = csv.Sniffer().sniff(str.join('', sample_lines))
                    self.log.debug("Detected CSV delimiter '%s'", csv_dialect.delimiter)
                except csv.Error:
                    # in case of any auto-detection failure, fall back to the default
                    self.log.warning("Could not determine field delimiter in file '%s',"
                                     " assuming it's a comma character (',').",
                                     path)
                    csv_dialect = 'excel'
                inputfile.seek(0)
                for lineno, row in enumerate(csv.reader(inputfile, csv_dialect)):
                    # ignore blank and comment lines (those that start with '#')
                    if len(row) == 0 or row[0].startswith('#'):
                        continue
                    try:
                        (iterno, N_str, p_mut_coeff_str, choose_or_rand_str, sick_or_not_str, off_v_last_str) = row
                    except ValueError:
                        self.log.error("Wrong format in line %d of file '%s':"
                                       " need 6 comma-separated values, but only got %d ('%s')."
                                       " Ignoring input line, fix it and re-run.",
                                       lineno+1, path, len(row), str.join(',', (str(x) for x in row)))
                        continue # with next `row`
                    # extract parameter values
                    try:
                        iterno = int(iterno)
                        N = GMhcCoevScript._parse_N(N_str)
                        p_mut_coeff = GMhcCoevScript._parse_p_mut_coeff(p_mut_coeff_str)
                        choose_or_rand = GMhcCoevScript._parse_choose_or_rand(choose_or_rand_str)
                        sick_or_not = GMhcCoevScript._parse_sick_or_not(sick_or_not_str)
                        off_v_last = GMhcCoevScript._parse_off_v_last(off_v_last_str)
                    except ValueError, ex:
                        self.log.warning("Ignoring line '%s' in input file '%s': %s",
                                         str.join(',', row), path, str(ex))
                        continue
                    basename = ('MHC_coev_' +
                                GMhcCoevScript._params_to_str(
                                    N, p_mut_coeff, choose_or_rand,
                                    sick_or_not, off_v_last))

                    # prepare job(s) to submit
                    if (iterno > iters[basename]):
                        self.log.info(
                                "Requested %d iterations for %s: %d already in session, preparing %d more",
                                iterno, basename, iters[basename], iterno - iters[basename])
                        for iter in range(iters[basename]+1, iterno+1):
                            kwargs = extra.copy()
                            base_output_dir = kwargs.pop('output_dir', self.params.output)
                            yield GMhcCoevTask(
                                self.params.walltime,    # single_run_duration
                                self.params.generations,
                                N,
                                p_mut_coeff,
                                choose_or_rand,
                                sick_or_not,
                                off_v_last,
                                executable=self.params.executable,
                                jobname=('%s#%d' % (basename, iter)),
                                output_dir = os.path.join(
                                    base_output_dir,
                                    ("N%d" % N),    # population size
                                    basename,       # params,
                                    str(iter),      # this iteration nr.
                                    ),
                                **kwargs)

            else:
                self.log.error("Ignoring input file '%s': not a CSV file.", path)


    ##
    ## INTERNAL METHODS
    ##


    _P_MUT_COEFF_RE = re.compile(r'((?P<num>[0-9]+)x)?10min(?P<exp>[0-9]+)')

    @staticmethod
    def _parse_p_mut_coeff(p_mut_coeff_str):
        try:
            return float(p_mut_coeff_str)
        except ValueError:
            match = GMhcCoevScript._P_MUT_COEFF_RE.match(p_mut_coeff_str)
            if not match:
                raise ValueError("Cannot parse P_MUT_COEFF expression '%s'" % p_mut_coeff_str)
            return float(match.group('num')) * 10.0**(-int(match.group('exp')))

    _N_RE = re.compile(r'N(?P<N>[0-9]+)')

    @staticmethod
    def _parse_N(N_str):
        try:
            return int(N_str)
        except ValueError:
            match = GMhcCoevScript._N_RE.match(N_str)
            if not match:
                raise ValueError("Cannot parse N expression '%s'" % N_str)
            return int(match.group('N'))

    @staticmethod
    def _parse_choose_or_rand(choose_or_rand_str):
        if choose_or_rand_str == "RM":
            return 1
        elif choose_or_rand_str == "DMAM":
            return 2
        elif choose_or_rand_str == "DMSSGD":
            return 3
        else:
            raise ValueError("Cannot parse CHOOSE_OR_RAND expression '%s'" % choose_or_rand_str)

    @staticmethod
    def _parse_sick_or_not(sick_or_not):
        if sick_or_not == "pat_on":
            return 1
        elif sick_or_not == "pat_off":
            return 0
        else:
            raise ValueError("Cannot parse SICK_OR_NOT expression '%s'" % sick_or_not)

    @staticmethod
    def _parse_off_v_last(off_v_last):
        try:
            return float(off_v_last)
        except ValueError:
            if not off_v_last.startswith("offval_"):
                raise ValueError("Cannot parse OFF_V_LAST expression '%s'" % off_v_last)
            off_v_last = off_v_last[7:]
            if off_v_last.startswith("0"):
                return float("0." + off_v_last[1:])
            elif off_v_last.startswith("1"):
                return 1.0
            else:
                raise ValueError("Cannot parse OFF_V_LAST expression '%s'" % off_v_last)

    @staticmethod
    def _string_to_params(s):
        """
        Return a tuple (N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last)
        obtained by parsing the string `s`.
        """
        p_mut_coeff, N, choose_or_rand, sick_or_not, off_v_last = s.split('__')

        return (GMhcCoevScript._parse_N(N),
                GMhcCoevScript._parse_p_mut_coeff(p_mut_coeff),
                GMhcCoevScript._parse_choose_or_rand(choose_or_rand),
                GMhcCoevScript._parse_sick_or_not(sick_or_not),
                GMhcCoevScript._parse_off_v_last(off_v_last))

    @staticmethod
    def _N_to_str(N):
        return ("N%d" % N)

    @staticmethod
    def _p_mut_coeff_to_str(p_mut_coeff):
        """
        Print the floating-point number `p_mut_coeff` as ``1x10min3``
        instead of the usual ``1e-3``.

        Examples::

          >>> _p_mut_coeff_to_str(0.005)
          '5x10min3'
          >>> _p_mut_coeff_to_str(10)
          '1x10plus1'
          >>> _p_mut_coeff_to_str(0.1234)
          '1.234x10min1'

        """
        exponent = math.log10(p_mut_coeff)
        mantissa = p_mut_coeff * (10 ** math.ceil(-exponent))
        # format output string
        if (mantissa == int(mantissa)):
            # no fractional part
            result = str(int(mantissa))
        else:
            result = str(mantissa)
        if exponent != 0:
            if exponent < 0:
                result += ('x10min%d' % (-int(exponent),))
            else:
                result += ('x10plus%d' % (int(exponent),))
        return result

    @staticmethod
    def _choose_or_rand_to_str(choose_or_rand):
        if choose_or_rand == 1:
            return "RM"
        elif choose_or_rand == 2:
            return "DMAM"
        elif choose_or_rand == 3:
            return "DMSSGD"
        else:
            raise ValueError("Valid values for `choose_or_rand` are: 1,2,3;"
                             " got '%s' instead" % choose_or_rand)

    @staticmethod
    def _sick_or_not_to_str(sick_or_not):
        if sick_or_not:
            return "pat_on"
        else:
            return "pat_off"

    @staticmethod
    def _off_v_last_to_str(off_v_last):
        if off_v_last < 0.0 or off_v_last > 1.0:
            raise ValueError("Parameter `off_v_last` must be in range 0.0 to 1.0;"
                             " got `%s`instead." % off_v_last)
        if off_v_last == 1.0:
            return "offval_1"
        else:
            # off_v_last == 0.xxxx

            # XXX: Python switches to scientific notation for
            # N<0.00001; the funny %-hack below ensures that the
            # dot+digits notation is used instead; starting Python
            # 2.6, the `format` function is available and should be
            # preferred:: off_v_last_str = format(off_v_last, 'f')
            off_v_last_str = ("%.16f" % off_v_last).rstrip('0')
            return ('offval_0' + off_v_last_str[2:])

    @staticmethod
    def _params_to_str(N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last):
        return str.join('__', [
            GMhcCoevScript._p_mut_coeff_to_str(p_mut_coeff),
            GMhcCoevScript._N_to_str(N),
            GMhcCoevScript._choose_or_rand_to_str(choose_or_rand),
            GMhcCoevScript._sick_or_not_to_str(sick_or_not),
            GMhcCoevScript._off_v_last_to_str(off_v_last),
            ])



#! /usr/bin/env python
#
"""
Specialized support for computational jobs running GAMESS-US.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'


import gc3libs
import gc3libs.application
import gc3libs.application.apppot
import gc3libs.exceptions
import os
import os.path
import re


class GamessApplication(gc3libs.Application):
    """
    Specialized `Application` object to submit computational jobs running GAMESS-US.

    The only required parameter for construction is the input file
    name; subsequent positional arguments are additional input files,
    that are added to the `Application.inputs` list, but not otherwise
    treated specially.

    The `verno` parameter is used to request a specific version of
    GAMESS-US; if the default value ``None`` is used, the default
    version of GAMESS-US at the executing site is run.

    Any other keyword parameter that is valid in the `Application`
    class can be used here as well, with the exception of `input` and
    `output`.  Note that a GAMESS-US job is *always* submitted with
    `join = True`, therefore any `stderr` setting is ignored.
    """

    application_name = 'gamess'

    def __init__(self, inp_file_path, *other_input_files, **extra_args):
        input_file_name = os.path.basename(inp_file_path)
        input_file_name_sans = os.path.splitext(input_file_name)[0]
        output_file_name = input_file_name_sans + '.dat'
        # add defaults to `extra_args` if not already present
        extra_args.setdefault('stdout', input_file_name_sans + '.out')
        extra_args.setdefault('tags', list())
        self.verno = extra_args.get('verno', None)
        if self.verno is not None:
            extra_args['tags'].append("APPS/CHEM/GAMESS-%s" % self.verno)
        else:
            # no VERNO specified
            extra_args['tags'].append("APPS/CHEM/GAMESS")
        # `rungms` has a fixed structure for positional arguments:
        # INPUT VERNO NCPUS; if one of them is to be omitted,
        # we use the empty string instead.
        arguments = [
            "rungms",
            input_file_name,
            str(extra_args.get('verno') or ""),
            str(extra_args.get('requested_cores') or "")
            ]
        if 'extbas' in extra_args:
           other_input_files += extra_args['extbas']
           arguments.extend(['--extbas', os.path.basename(extra_args['extbas'])])
        #set job name
        extra_args['jobname'] = input_file_name_sans
        # issue WARNING about memory handling
        if 'requested_memory' in extra_args:
            # XXX: should this be an error instead?
            gc3libs.log.warning(
                "Requested %s of memory per core; depending on how the execution site"
                " handles memory limits, this may lead to an error in the ``ddikick``"
                " startup.  In that case, re-run without memory specification.",
                extra_args['requested_memory'])
        # build generic `Application` obj
        gc3libs.Application.__init__(self,
                                     arguments = arguments,
                                     inputs = [ inp_file_path ] + list(other_input_files),
                                     outputs = [ output_file_name ],
                                     join = True,
                                     # needed by `ggamess`
                                     inp_file_path = inp_file_path,
                                     **extra_args)

    _termination_re = re.compile(r'EXECUTION \s+ OF \s+ GAMESS \s+ TERMINATED \s+-?(?P<gamess_outcome>NORMALLY|ABNORMALLY)-?'
                                 r'|ddikick.x: .+ (exited|quit) \s+ (?P<ddikick_outcome>gracefully|unexpectedly)', re.X)

    def terminated(self):
        """
        Append to log the termination status line as extracted from
        the GAMESS '.out' file.

        The job exit code `.execution.exitcode` is (re)set according
        to the following table:

        =====================  ===============================================
        Exit code              Meaning
        =====================  ===============================================
        0                      the output file contains the string \
                               ``EXECUTION OF GAMESS TERMINATED normally``
        1                      the output file contains the string \
                               ``EXECUTION OF GAMESS TERMINATED -ABNORMALLY-``
        2                      the output file contains the string \
                               ``ddikick exited unexpectedly``
        70 (`os.EX_SOFTWARE`)  the output file cannot be read or \
                               does not match any of the above patterns
        =====================  ===============================================

        """
        gc3libs.log.debug("Running GamessApplication post-processing hook...")
        output_dir = self.output_dir
        if self.stdout:
            output_filename = os.path.join(output_dir, self.stdout)
        else:
            output_filename = os.path.join(
                output_dir,
                os.path.splitext(os.path.basename(self.inp_file_path))[0] + '.out')
        if not os.path.exists(output_filename):
            # no output file, override exit code if it indicates success
            if self.execution.exitcode == os.EX_OK:
                self.execution.exitcode = os.EX_SOFTWARE
        else:
            # output file exists, start with pessimistic default and
            # override if we find a "success" keyword
            self.execution.exitcode = os.EX_SOFTWARE # internal software error
            gc3libs.log.debug("Trying to read GAMESS termination status"
                              " off output file '%s' ..." % output_filename)
            output_file = open(output_filename, 'r')
            for line in output_file:
                match = self._termination_re.search(line)
                if match:
                    if match.group('gamess_outcome'):
                        self.execution.info = line.strip().capitalize()
                        if match.group('gamess_outcome') == 'ABNORMALLY':
                            self.execution.exitcode = 1
                        elif match.group('gamess_outcome') == 'NORMALLY':
                            self.execution.exitcode = 0
                        break
                    elif match.group('ddikick_outcome'):
                        self.execution.info = line.strip().capitalize()
                        if match.group('ddikick_outcome') == 'unexpectedly':
                            self.execution.exitcode = 2
                        # If the GAMESS execution terminated normally, this would
                        # have been catched by the test above; otherwise we risk here
                        # setting the exitcode to EX_OK even if GAMESS didn't run at
                        # all, but `ddikick` does not flag it as an error. (Not sure
                        # this can actually happen, but better err on the safe side.)
                        #elif match.group('ddikick_outcome') == 'gracefully':
                        #    gc3libs.log.info('Setting exit code to 0')
                        #    self.execution.exitcode = 0
                        break
                    else:
                        raise AssertionError(
                            "Input line '%s' matched,"
                            " but neither group 'gamess_outcome'"
                            " nor 'ddikick_outcome' did!")
            output_file.close()



class GamessAppPotApplication(GamessApplication,
                              gc3libs.application.apppot.AppPotApplication):
    """
    Specialized `AppPotApplication` object to submit computational
    jobs running GAMESS-US.

    This class makes no check or guarantee that a GAMESS-US executable
    will be available in the executing AppPot instance: the
    `apppot_img` and `apppot_tag` keyword arguments can be used to
    select the AppPot system image to run this application; see the
    `AppPotApplication`:class: for information.

    The `__init__` construction interface is compatible with the one
    used in `GamessApplication`:class:. The only required parameter for
    construction is the input file name; any other argument names an
    additional input file, that is added to the `Application.inputs`
    list, but not otherwise treated specially.

    Any other keyword parameter that is valid in the `Application`
    class can be used here as well, with the exception of `input` and
    `output`.  Note that a GAMESS-US job is *always* submitted with
    `join = True`, therefore any `stderr` setting is ignored.
    """

    application_name = 'gamess'

    def __init__(self, inp_file_path, *other_input_files, **extra_args):
        input_file_name = os.path.basename(inp_file_path)
        input_file_name_sans = os.path.splitext(input_file_name)[0]
        output_file_name = input_file_name_sans + '.dat'
        # add defaults to `extra_args` if not already present
        extra_args.setdefault('stdout', input_file_name_sans + '.out')
        extra_args.setdefault('application_tag', "gamess")
        extra_args.setdefault('tags', list())
        extra_args['tags'].insert(0, "APPS/CHEM/GAMESS-APPPOT-0.11.11.08")
        if extra_args.has_key('extbas') and extra_args['extbas'] is not None:
            other_input_files += extra_args['extbas']
            arguments.extend(['--extbas', os.path.basename(extbas)])
        # set job name
        extra_args['jobname'] = input_file_name_sans
        # init superclass
        gc3libs.application.apppot.AppPotApplication.__init__(
            self,
            # `rungms` has a fixed structure for positional arguments:
            # INPUT VERNO NCPUS; if one of them is to be omitted,
            # we cannot use the empty string instead because `apppot-start`
            # cannot detect it from the kernel command line, so we have to
            # hard-code default values here...
            arguments = [
                "localgms",
                input_file_name,
                str(extra_args.get('verno') or "00"),
                str(extra_args.get('requested_cores') or "")
                ],
            inputs = [ inp_file_path ] + list(other_input_files),
            outputs = [ output_file_name ],
            join = True,
            # needed by `ggamess`
            inp_file_path = inp_file_path,
            **extra_args)


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="gamess",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
Deal with GC3Pie configuration files.
"""
# Copyright (C) 2012, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'


# stdlib imports
import ConfigParser
import inspect
import os
import re
import sys

# GC3Pie imports
import gc3libs
import gc3libs.authentication
from gc3libs.compat._collections import defaultdict
import gc3libs.utils

from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.utils import defproperty


## auxiliary methods for `Configuration`
#
# these must be defined before `Configuration` is parsed, because they
# are referenced in the definition of `Configuration` itself
#

# map values for the `architecture=...` configuration item
# into internal constants
_architecture_value_map = {
    # 'x86-32', 'x86 32-bit', '32-bit x86' and variants thereof
    re.compile('x86[ _-]+32([ _-]*bits?)?', re.I): gc3libs.Run.Arch.X86_32,
    re.compile('32[ _-]*bits? +[ix]86', re.I):     gc3libs.Run.Arch.X86_32,
    # accept also values printed by `uname -a` on 32-bit x86 archs
    re.compile('i[3456]86', re.I):                 gc3libs.Run.Arch.X86_32,
    # 'x86_64', 'x86 64-bit', '64-bit x86' and variants thereof
    re.compile('x86[ _-]+64([ _-]*bits?)?', re.I): gc3libs.Run.Arch.X86_64,
    re.compile('64[ _-]*bits? +[ix]86', re.I):     gc3libs.Run.Arch.X86_64,
    # also accept commercial arch names
    re.compile('(amd[ -]*64|x64|emt64|intel[ -]*64)( *bits?)?', re.I): \
                                                   gc3libs.Run.Arch.X86_64,
    # finally, map "32-bit" and "64-bit" to i686 and x86_64
    re.compile('32[ _-]*bits?', re.I):             gc3libs.Run.Arch.X86_32,
    re.compile('64[ _-]*bits?', re.I):             gc3libs.Run.Arch.X86_64,
    }

def _parse_architecture(arch_str):
    def matching_architecture(value):
        for matcher, arch in _architecture_value_map.iteritems():
            if matcher.match(value):
                return arch
        raise ValueError("Unknown architecture '%s'." % value)
    archs = [ matching_architecture(value.strip())
              for value in arch_str.split(',') ]
    if len(archs) == 0:
        raise ValueError("Empty or invalid 'architecture' setting.")
    return set(archs)

def _legacy_parse_duration(duration_str):
    try:
        # old-style config: integral number of hours
        return int(duration_str)*hours
    except ValueError:
        # apply `Duration` parsing rules; if this fails, users will
        # see the error message from the `Duration` parser.
        return Duration(duration_str)

def _legacy_parse_memory(memory_str):
    try:
        # old-style config: integral number of GBs
        return int(memory_str)*GB
    except ValueError:
        # apply usual quantity parsing rules; if this fails, users
        # will see the error message from the `Memory`/`Quantity` parser.
        return Memory(memory_str)


## the main class of this module

class Configuration(gc3libs.utils.Struct):
    """
    In-memory representation of the GC3Pie configuration.

    This class provides facilities for:

    * parsing configuration files (methods `load`:meth: and
      `merge_file`:meth:);
    * validating the loaded values;
    * instanciating the internal GC3Pie objects resulting from the
      configuration (methods `make_auth`:meth: and
      `make_resource`:meth:).

    The constructor takes a list of files to load (`locations`) and a
    list of key=value pairs to provide defaults for the configuration.
    Both lists are optional and can be omitted, resulting in a
    configuration containing only GC3Pie default values.

    Example 1: initialization from config file::

      >>> import os
      >>> example_cfgfile = os.path.join(os.path.dirname(__file__), 'etc/gc3pie.conf.example')
      >>> cfg = Configuration(example_cfgfile)
      >>> cfg.debug
      '0'

    Example 2: initialization from key=value list::

      >>> cfg = Configuration(auto_enable_auth=False, foo=1, bar='baz')
      >>> cfg.auto_enable_auth
      False
      >>> cfg.foo
      1
      >>> cfg.bar
      'baz'

    When both a configuration file *and* a key=value list is present,
    values in the configuration files override those in the key=value
    list::

      >>> cfg = Configuration(example_cfgfile, debug=1)
      >>> cfg.debug
      '0'

    Example 3: default initialization::

      >>> cfg = Configuration()
      >>> cfg.auto_enable_auth
      True

    """

    def __init__(self, *locations, **extra_args):
        self._auth_factory = None

        # these fields are required
        self.resources = defaultdict(gc3libs.utils.Struct)
        self.auths = defaultdict(gc3libs.utils.Struct)

        # use keyword arguments to set defaults
        self.auto_enable_auth = extra_args.pop('auto_enable_auth', True)
        self.update(extra_args)

        # save the list of (valid) config files
        self.cfgfiles = []

        # load configuration files if any
        if len(locations) > 0:
            self.load(*locations)


    def load(self, *locations):
        """
        Merge settings from configuration files into this `Configuration` instance.

        Environment variables and `~` references are expanded in the
        location file names.

        If any of the specified files does not exist or cannot be read
        (for whatever reason), a message is logged but the error is
        ignored.  However, a `NoConfigurationFile` exception is raised
        if *none* of the specified locations could be read.

        :raise gc3libs.exceptions.NoConfigurationFile:
            if none of the specified files could be read.
        """
        files_successfully_read = 0

        for filename in locations:
            filename = os.path.expandvars(os.path.expanduser(filename))
            if os.path.exists(filename):
                if not os.access(filename, os.R_OK):
                    gc3libs.log.debug("Configuration.load(): File '%s' cannot be read, ignoring." % filename)
                    continue # with next `filename`
            else:
                gc3libs.log.debug("Configuration.load(): File '%s' does not exist, ignoring." % filename)
                continue # with next `filename`

            try:
                self.cfgfiles.append(os.path.abspath(filename))
                self.merge_file(filename)
                files_successfully_read += 1
            except gc3libs.exceptions.ConfigurationError:
                continue # with next file

        if files_successfully_read == 0:
            raise gc3libs.exceptions.NoConfigurationFile(
                "Could not read any configuration file; tried location '%s'."
                % str.join("', '", locations))


    def merge_file(self, filename):
        """
        Read configuration files and merge the settings into this `Configuration` object.

        Contrary to `load`:meth: (which see), the file name is taken
        literally and an error is raised if the file cannot be read
        for whatever reason.

        Any parameter which is set in the configuration files
        ``[DEFAULT]`` section, and whose name does not start with
        underscore (``_``) defines an attribute in the current
        `Configuration`.

        .. warning::

          No type conversion is performed on values set this way - so
          they all end up being strings!

        :raise gc3libs.exceptions.ConfigurationError: if the
            configuration file does not exist, cannot be read, is
            corrupt or has wrong format.
        """
        gc3libs.log.debug("Configuration.load(): Reading file '%s' ..." % filename)
        stream = open(filename, 'r')
        (defaults, resources, auths) = self._parse(stream, filename)
        stream.close()
        for name, values in resources.iteritems():
            self.resources[name].update(values)
        for name, values in auths.iteritems():
            self.auths[name].update(values)
        for name, value in defaults.iteritems():
            if not name.startswith('_'):
                self[name] = value


    def _parse(self, stream, filename=None):
        """
        Read configuration file and return a `(defaults, resources, auths)` triple.

        The members of the result triple are as follows:

        * `defaults`: a dictionary containing keys found in the
          ``[DEFAULTS]`` section of the configuration file (if any);

        * `resources`: a dictionary mapping resource names into a
          dictionary of key/value attributes contained in the
          configuration file under the ``[resource/name]`` heading;

        * `auths`: same for the ``[auth/name]`` sections.

        In addition, key renaming (for compatibility with previous versions) and
        type conversion is performed here, so that the returned dictionaries
        conform to a specified schema:

          ===================  =========================
          Attribute name       Type
          ===================  =========================
          architecture         set of strings
          max_cores            int
          max_cores_per_job    int
          max_memory_per_core  gc3libs.quantity.Memory
          max_walltime         gc3libs.quantity.Duration
          ===================  =========================

        Any attribute not mentioned in the above table will have type
        ``str`` (i.e., it is left unchanged).

        """
        defaults = dict()
        resources = defaultdict(dict)
        auths = defaultdict(dict)

        parser = ConfigParser.SafeConfigParser()
        try:
            parser.readfp(stream, filename)
        except ConfigParser.Error, err:
            if filename is None:
                if hasattr(stream, 'name'):
                    filename = stream.name
                else:
                    filename = repr(stream)
            raise gc3libs.exceptions.ConfigurationError(
                "Configuration file '%s' is unreadable or malformed: %s: %s"
                % (filename, err.__class__.__name__, str(err)))

        # update `defaults` with the contents of the `[DEFAULTS]` section
        defaults.update(parser.defaults())

        for sectname in parser.sections():
            if sectname.startswith('auth/'):
                # handle auth section
                name = sectname.split('/', 1)[1]
                gc3libs.log.debug("Config._parse():"
                                  " Read configuration stanza for auth '%s'." % name)
                config_items = dict(parser.items(sectname))
                auths[name].update(config_items)
                auths[name]['name'] = name
                if __debug__:
                    gc3libs.log.debug(
                        "Config._parse(): Auth '%s' defined by: %s.",
                        name, str.join(', ', [
                            ("%s=%r" % (k,v)) for k,v in sorted(auths[name].iteritems())
                            ]))

            elif  sectname.startswith('resource/'):
                # handle resource section
                name = sectname.split('/', 1)[1]
                gc3libs.log.debug("Config._parse():"
                                  " Read configuration stanza for resource '%s'." % name)

                config_items = dict(parser.items(sectname))
                self._perform_key_renames(config_items, self._renamed_keys, filename)
                self._perform_value_updates(config_items, self._updated_values, filename)
                self._perform_filename_conversion(config_items, self._path_key_regexp, filename)
                try:
                    self._perform_type_conversions(config_items, self._convert, filename)
                except Exception, err:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Incorrect entry for resource '%s' in configuration file '%s': %s"
                        % (name, filename, str(err)))

                resources[name].update(config_items)
                resources[name]['name'] = name
                if __debug__:
                    gc3libs.log.debug(
                        "Config._parse(): Resource '%s' defined by: %s.",
                        name, str.join(', ', [
                            ("%s=%r" % (k,v)) for k,v in sorted(resources[name].iteritems())
                            ]))

            else:
                # Unhandled sectname
                gc3libs.log.warning(
                    "Config._parse(): unknown configuration section '%s' -- ignoring!",
                    sectname)

        return (defaults, resources, auths)

    _renamed_keys = {
        # old key name         new key name
        # ===================  ===================
        'ncores':              'max_cores',
        'sge_accounting_delay':'accounting_delay',
        }

    @staticmethod
    def _perform_key_renames(config_items, renames, filename):
        for oldkey, newkey in renames.iteritems():
            if oldkey in config_items:
                gc3libs.log.warning(
                    "Configuration item '%s' was renamed to '%s',"
                    " please change occurrences of '%s' to '%s'"
                    " in configuration file '%s'.",
                    oldkey, newkey, oldkey, newkey, filename)
                if newkey in config_items:
                    # drop
                    gc3libs.log.error(
                        "Both old-style configuration item '%s' and new-style '%s'"
                        " detected in file '%s': ignoring old-style item '%s=%s'.",
                        oldkey, newkey, filename, config_items[oldkey])
                else:
                    config_items[newkey] = config_items[oldkey]
                del config_items[oldkey]

    _updated_values = {
        # key name  old value            new value
        # ========  ===================  ==================
        'type': {
            # old value     new value
            # ============  ==================
            'arc':          gc3libs.Default.ARC0_LRMS,
            'ssh':          gc3libs.Default.SGE_LRMS,
            'subprocess':   gc3libs.Default.SHELLCMD_LRMS,
            },
        }

    @staticmethod
    def _perform_value_updates(config_items, renames, filename):
        for key, changed in renames.iteritems():
            if key in config_items:
                value = config_items[key]
                if value in changed:
                    gc3libs.log.warning(
                        "Configuration value '%s' was renamed to '%s',"
                        " please change occurrences of '%s=%s' to '%s=%s'"
                        " in configuration file '%s'.",
                        value, changed[value],
                        key, value, key, changed[value],
                        filename)
                    config_items[key] = changed[value]

    # type transforms for well-known configuration keys
    _convert = {
        # item name            converter
        # ===================  ==================================
        'enabled':             gc3libs.utils.string_to_boolean,
        'architecture':        _parse_architecture,
        'max_cores':           int,
        'max_cores_per_job':   int,
        'max_memory_per_core': _legacy_parse_memory,
        'max_walltime':        _legacy_parse_duration,
        }


    @staticmethod
    def _perform_filename_conversion(config_items, path_regexp, filename):
        for key, value in config_items.iteritems():
            if path_regexp.match(key):
                config_items[key] = os.path.join(
                    os.path.dirname(filename),
                    value)

    _path_key_regexp = re.compile('^(\w+_)?(prologue|epilogue)')

    @staticmethod
    def _perform_type_conversions(config_items, converters, filename):
        for key, converter in converters.iteritems():
            if key in config_items:
                try:
                    config_items[key] = converter(config_items[key])
                except Exception, err:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Error parsing configuration item '%s': %s: %s"
                        % (key, err.__class__.__name__, str(err)))


    @defproperty
    def auth_factory():
        """
        The instance of `gc3libs.authentication.Auth`:class: used to
        manage auth access for the resources.

        This is a *read-only* attribute, created upon first access
        with the values set in `self.auths` and `self.auto_enabled`.
        """
        def fget(self):
            if self._auth_factory is None:
                try:
                    self._auth_factory = gc3libs.authentication.Auth(self.auths, self.auto_enable_auth)
                except Exception, err:
                    gc3libs.log.critical(
                        "Failed initializing Auth module: %s: %s",
                        err.__class__.__name__, str(err))
                    raise
            return self._auth_factory
        return locals()


    def make_auth(self, name):
        """
        Return factory for auth credentials configured in section ``[auth/name]``.
        """
        # use `lambda` for delayed evaluation
        return (lambda **extra_args: self.auth_factory.get(name, **extra_args))


    def make_resources(self, ignore_errors=True):
        """
        Make backend objects corresponding to the configured resources.

        Return a dictionary, mapping the resource name (string) into
        the corresponding backend object.

        By default, errors in constructing backends (e.g., due to a
        bad configuration) are silently ignored: the offending
        configuration is just dropped.  This can be changed by setting
        the optional argument `ignore_errors` to `False`: in this
        case, an exception is raised whenever we fail to construct a
        backend.
        """
        resources = { }
        for name, resdict in self.resources.iteritems():
            try:
                backend = self._make_resource(resdict)
                if backend is None: # resource is disabled
                    continue
                assert name == backend.name
            except Exception, err:
                # Print the backtrace only if loglevel is DEBUG or
                # more.
                exc_info = gc3libs.log.level <= gc3libs.logging.DEBUG
                gc3libs.log.warning(
                    "Failed creating backend for resource '%s' of type '%s': %s: %s",
                    resdict.get('name', '(unknown name)'),
                    resdict.get('type', '(unknown type)'),
                    err.__class__.__name__, str(err), exc_info=exc_info)
                if ignore_errors:
                    continue
                else:
                    raise
            resources[name] = backend
        return resources


    def _make_resource(self, resdict):
        """
        Return a backend initialized from the key/value pairs in `resdict`.
        """
        # name' should have been defined by the caller, so if it's
        # missing it's an internal coherency error
        assert 'name' in resdict, (
            "Invalid resource definition '%s': missing required key 'name'."
            % resdict)

        # default values
        resdict.setdefault('enabled', True)
        if not resdict['enabled']:
            gc3libs.log.info(
                "Dropping computational resource '%s'"
                " because of 'enabled=False' setting"
                " in configuration file.",
                resdict['name'])
            return None

        if __debug__:
            gc3libs.log.debug(
                "Creating resource '%s' defined by: %s.",
                resdict['name'], str.join(', ', [
                    ("%s=%r" % (k,v)) for k,v in sorted(resdict.iteritems())
                    ]))

        # sanity check
        if 'type' not in resdict:
            raise gc3libs.exceptions.ConfigurationError(
                "Missing required parameter 'type' in resource definition %s."
                % resdict)

        # XXX: should be done by the backend constructor!?
        if 'architecture' not in resdict:
            raise gc3libs.exceptions.ConfigurationError(
                "No architecture specified for resource '%s'"
                % resdict['name'])

        if 'auth' in resdict:
            resdict['auth'] = self.make_auth(resdict['auth'])

        try:
            if resdict['type'] == gc3libs.Default.ARC0_LRMS:
                from gc3libs.backends.arc0 import ArcLrms
                cls = ArcLrms
            elif resdict['type'] == gc3libs.Default.ARC1_LRMS:
                from gc3libs.backends.arc1 import Arc1Lrms
                cls = Arc1Lrms
            elif resdict['type'] == gc3libs.Default.SGE_LRMS:
                from gc3libs.backends.sge import SgeLrms
                cls = SgeLrms
            elif resdict['type'] == gc3libs.Default.PBS_LRMS:
                from gc3libs.backends.pbs import PbsLrms
                cls = PbsLrms
            elif resdict['type'] == gc3libs.Default.LSF_LRMS:
                from gc3libs.backends.lsf import LsfLrms
                cls = LsfLrms
            elif resdict['type'] == gc3libs.Default.SHELLCMD_LRMS:
                from gc3libs.backends.shellcmd import ShellcmdLrms
                cls = ShellcmdLrms
            elif resdict['type'] == gc3libs.Default.SLURM_LRMS:
                from gc3libs.backends.slurm import SlurmLrms
                cls = SlurmLrms
            elif resdict['type'].split('+')[0] == gc3libs.Default.EC2_LRMS:
                from gc3libs.backends.ec2 import EC2Lrms
                cls = EC2Lrms
            else:
                raise gc3libs.exceptions.ConfigurationError(
                    "Unknown resource type '%s'" % resdict['type'])
            # check that required parameters are given, and try to
            # give a sensible error message if not; if we do not
            # do this, users see a traceback like this::
            #
            #   gc3.gc3libs: ERROR: Could not create resource 'schroedinger-via-ssh': __init__() takes at least 10 non-keyword arguments (9 given). Configuration file problem?
            #
            # which gives no clue about what to correct!
            args, varargs, keywords, defaults = inspect.getargspec(cls.__init__)
            if defaults is not None:
                # `defaults` is a list of default values for the last N args
                defaulted = dict((argname, value)
                                for argname, value in zip(reversed(args),
                                                          reversed(defaults)))
            else:
                # no default values at all
                defaulted = { }
            for argname in args[1:]: # skip `self`
                if argname not in resdict and argname not in defaulted:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Missing required configuration parameter '%s' for resource '%s'"
                        % (argname, resdict['name']))


            # finally, try to construct backend class...
            return cls(**dict(resdict))
        except Exception, err:
            gc3libs.log.error(
                "Could not create resource '%s': %s. Configuration file problem?"
                % (resdict['name'], str(err)))
            raise


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="config",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Riccardo Murri <riccardo.murri@gmail.com>'
# summary of user-visible changes
__changelog__ = """
  2012-11-06:
    * Initial draft.
"""
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import gsparsegrid
    gsparsegrid.SparseGridScript().run()


# stdlib imports
import itertools
import os
import os.path
import string
import sys


## interface to Gc3libs
import gc3libs
from gc3libs import Application, Run, Task
import gc3libs.exceptions
from gc3libs.cmdline import SessionBasedScript, executable_file, nonnegative_int
import gc3libs.utils



class SparseGridApplication(Application):
    application_name = 'sgrid'
    def __init__(self, np, total_memory, fnum, dimension, lmax, epsilon,
                 executable=None, **extra_args):
        """
        Execute one instance of the 'sparse grid' application.

        Parameters are as follows:

        :param int   np:        Number of processors to use
        :param int   fnum:      Test function number [1-6]
        :param int   dimension: 1 -120
        :param int   Lmax:
        :param float epsilon: e.g., 1e-3, 1e-4, 1e-5
        """
        extra_args['requested_cores'] = np
        extra_args['requested_memory'] = total_memory

        # build Application object
        gc3libs.Application.__init__(
            self,
            executable = '',
            arguments = [
                # std MPI startup
                'mpiexec', '-n', np,
                # actual `sgrid` invocation
                "$SGRID_EXEC", fnum, dimension, lmax, epsilon
                ],
            inputs = [], # no input files
            outputs = [], # no output files (except stdout/stderr)
            join = True,
            stdout = 'sgrid.log',
            tags = [ 'TEST/SGRID-0.1' ],
            **extra_args
            )


class SparseGridScript(SessionBasedScript):
    """
    Sample parameter-study script.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            )


    def setup_options(self):
        self.add_param("--fnum", action="store", dest="fnum",default='1',
                       help="Select function no. to test [1-6].")
        self.add_param("--dimension", action="store", dest="dimension",default='2',
                       help="Grid dimension.")
        self.add_param("--lmax", action="store", dest="lmax",default='3',
                       help="Select Lmax")
        self.add_param("--epsilon", action="store", dest="epsilon",default='1.e-3',
                       help="Select epsilon.")
        self.add_param("--mpi_cores", action="store", dest="mpi_cores",default='10',
                       help="Select number of cores to use. ")


    def new_tasks(self, extra):

        fnums = [ int(fnum) for fnum in self.params.fnum.split(',') ]
        dims = [ int(dim) for dim in self.params.dimension.split(',') ]
        lmaxs = [ int(lmax) for lmax in self.params.lmax.split(',') ]
        epsilons = [ float(epsilon) for epsilon in self.params.epsilon.split(',') ]
        ncores = [ int(ncore) for ncore in self.params.mpi_cores.split(',') ]

        # create one task per combination of the input parameters
        for fnum, dim, lmax, epsilon,ncore in itertools.product(fnums, dims, lmaxs, epsilons, ncores):
            jobname = ("sgrid_fnum=%d_dim=%d_Lmax=%d_epsilon=%g_mpi_cores=%d" % (fnum, dim, lmax, epsilon,ncore))

            extra_args = extra.copy()
            extra_args['output_dir'] = os.path.join(
                self.make_directory_path(self.params.output, jobname), 'output')
            extra_args['jobname'] = jobname

            # passed to Application as `requested_cores` in SparseGridApplication
            np = ncore
            # Compute total memory needed for a job with np cores. 
            total_memory = self.params.memory_per_core * np
            yield SparseGridApplication(np, total_memory, fnum, dim, lmax, epsilon, **extra_args)



#!/usr/bin/python

# NG's default packages install arclib into /opt/nordugrid/lib/pythonX.Y/site-packages;
# add this anyway in case users did not set their PYTHONPATH correctly
import sys
sys.path.append('/opt/nordugrid/lib/python%d.%d/site-packages' 
                % sys.version_info[:2])



#! /usr/bin/env python
#
# Copyright (C) 2009-2013 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
"""
GC3Libs is a python package for controlling the life-cycle of a Grid
or batch computational job.

GC3Libs provides services for submitting computational jobs to Grids
and batch systems, controlling their execution, persisting job
information, and retrieving the final output.

GC3Libs takes an application-oriented approach to batch computing. A
generic :class:`Application` class provides the basic operations for
controlling remote computations, but different :class:`Application`
subclasses can expose adapted interfaces, focusing on the most
relevant aspects of the application being represented.

"""
__docformat__ = 'reStructuredText'

__version__ = '2.1.0rc2 version (SVN $Revision: 3585 $)'


import copy
import os
import os.path
import string
import sys
import time
import types
import subprocess
import shlex

import logging
import logging.config
log = logging.getLogger("gc3.gc3libs")

from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.compat._collections import OrderedDict

# this needs to be defined before we import other GC3Libs modules, as
# they may depend on it
class Default(object):
    """
    A namespace for all constants and default values used in the
    GC3Libs package.
    """
    RCDIR = os.path.join(os.path.expandvars('$HOME'), ".gc3")
    CONFIG_FILE_LOCATIONS = [
        # system-wide config file
        "/etc/gc3/gc3pie.conf",
        # virtualenv config file
        os.path.expandvars("$VIRTUAL_ENV/etc/gc3/gc3pie.conf"),
        # user-private config file: first look into `$GC3PIE_CONF`, and
        # fall-back to `~/.gc3/gc3pie.conf`
        os.environ.get('GC3PIE_CONF', os.path.join(RCDIR, "gc3pie.conf"))
        ]
    JOBS_DIR = os.path.join(RCDIR, "jobs")
    ARC0_LRMS = 'arc0'
    ARC1_LRMS = 'arc1'
    ARC2_LRMS = 'arc2'
    ARC_CACHE_TIME = 30 #: only update ARC resources status every this seconds
    ARC_LOST_JOB_TIMEOUT = 4*ARC_CACHE_TIME #: consider a submitted job lost if it does not show up in the information system after this duration
    ARC_JOBLIST_LOCATION = os.path.expandvars("$HOME/.arc/jobs.xml")

    SGE_LRMS = 'sge'
    PBS_LRMS = 'pbs'
    LSF_LRMS = 'lsf'
    SHELLCMD_LRMS = 'shellcmd'
    SLURM_LRMS = 'slurm'
    SUBPROCESS_LRMS = 'subprocess'
    EC2_LRMS = 'ec2'

    # Transport information
    SSH_PORT = 22
    SSH_CONNECT_TIMEOUT = 30

    # Proxy
    PROXY_VALIDITY_THRESHOLD = 600 #: Proxy validity threshold in seconds. If proxy is expiring before the thresold, it will be marked as to be renewed.

    ARC1_LOGFILE = os.path.join(RCDIR, "arc1.log")
    ARC_LOGFILE = os.path.join(RCDIR, "arc.log")
    ARC1_DEFAULT_SERVICE_TIMEOUT = 3 # max wait (seconds) for a service to respond; hopefully this impacts also LDAP searches

    PEEK_FILE_SIZE = 120 # expressed in bytes

    CERTIFICATE_AUTHORITIES_DIR = "/etc/grid-security/certificates"
    VOMS_DIR = "/etc/grid-security/vomsdir"

from gc3libs.exceptions import *
from gc3libs.persistence import Persistable
from gc3libs.url import UrlKeyDict, UrlValueDict
from gc3libs.utils import defproperty, deploy_configuration_file, Enum, History, Struct, safe_repr


class Task(Persistable, Struct):
    """
    Mix-in class implementing a facade for job control.

    A `Task` can be described as an "active" job, in the sense that
    all job control is done through methods on the `Task` instance
    itself; contrast this with operating on `Application` objects
    through a `Core` or `Engine` instance.

    The following pseudo-code is an example of the usage of the `Task`
    interface for controlling a job.  Assume that `GamessApplication` is
    inheriting from `Task` (it actually is)::

        t = GamessApplication(input_file)
        t.submit()
        # ... do other stuff
        t.update_state()
        # ... take decisions based on t.execution.state
        t.wait() # blocks until task is terminated

    Each `Task` object has an `execution` attribute: it is an instance
    of class :class:`Run`, initialized with a new instance of `Run`,
    and at any given time it reflects the current status of the
    associated remote job.  In particular, `execution.state` can be
    checked for the current task status.

    After successful initialization, a `Task` instance will have the
    following attributes:

    `changed`
      evaluates to `True` if the `Task` has been changed since last
      time it has been saved to persistent storage (see
      :mod:`gclibs.persistence`)

    `execution`
      a `Run` instance; its state attribute is initially set to ``NEW``.

    """

    def __init__(self, **extra_args):
        """
        Initialize a `Task` instance.

        The following attributes are defined on a valid `Task` instance:

        * `execution`: a `gc3libs.Run`:class: instance

        :param grid: A :class:`gc3libs.Engine` or
                     :class:`gc3libs.Core` instance, or anything
                     implementing the same interface.
        """
        Persistable.__init__(self, **extra_args)
        Struct.__init__(self, **extra_args)
        self.execution = Run(attach=self)
        # `_controller` and `_attached` are set by `attach()`/`detach()`
        self._attached = False
        self._controller = None
        self.changed = True

    # manipulate the "controller" interface used to control the associated job
    def attach(self, controller):
        """
        Use the given Grid interface for operations on the job
        associated with this task.
        """
        if self._controller != controller:
            if self._attached:
                self.detach()
            #gc3libs.log.debug("Attaching %s to %s" % (self, controller))
            controller.add(self)
            self._attached = True
            self._controller = controller

    # create a class-shared fake "controller" object, that just throws a
    # DetachedFromController exception when any of its methods is used.  We
    # use this as a safeguard for detached `Task` objects, in order to
    # get sensible error reporting.
    class __NoController(object):
        # XXX: this returns a function object for whatever `name`;
        # should be fine since a "controller" interface should just contain
        # methods, but one never knows...
        def __getattr__(self, name):
            def throw_error(*args, **kwargs):
                raise gc3libs.exceptions.DetachedFromControllerError(
                    "Task object is not attached to a controller.")
            return throw_error
    __no_controller = __NoController()

    def detach(self):
        """
        Remove any reference to the current grid interface.  After
        this, calling any method other than :meth:`attach` results in
        an exception :class:`TaskDetachedFromGridError` being thrown.
        """
        if self._attached:

            self._attached = False
            try:
                self._controller.remove(self)
            except:
                pass
            self._controller = Task.__no_controller


    # interface with pickle/gc3libs.persistence: do not save the
    # attached grid/engine/core as well: it definitely needs to be
    # saved separately.

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_controller'] = None
        state['_attached'] = None
        state['changed'] = False
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self.detach()


    # grid-level actions on this Task object are re-routed to the
    # grid/engine/core instance
    def submit(self, resubmit=False, **extra_args):
        """
        Start the computational job associated with this `Task` instance.
        """
        assert self._attached, ("Task.submit() called on detached task %s." % self)
        assert hasattr(self._controller, 'submit'), \
               ("Invalid `_controller` object '%s' in Task %s" % (self._controller, self))
        self._controller.submit(self, resubmit, **extra_args)


    def update_state(self, **extra_args):
        """
        In-place update of the execution state of the computational
        job associated with this `Task`.  After successful completion,
        `.execution.state` will contain the new state.

        After the job has reached the `TERMINATING` state, the following
        attributes are also set:

        `execution.duration`
          Time lapse from start to end of the job at the remote
          execution site, as a `gc3libs.quantity.Duration`:class: value.
          (This is also often referred to as the 'wall-clock time' or
          `walltime`:term: of the job.)

        `execution.max_used_memory`
          Maximum amount of RAM used during job execution, represented
          as a `gc3libs.quantity.Memory`:class: value.

        `execution.used_cpu_time`
          Total time (as a `gc3libs.quantity.Duration`:class: value) that the
          processors has been actively executing the job's code.

        The execution backend may set additional attributes; the exact
        name and format of these additional attributes is
        backend-specific.  However, you can easily identify the
        backend-specific attributes because their name is prefixed
        with the (lowercased) backend name; for instance, the
        `PbsLrms`:class: backend sets attributes `pbs_queue`,
        `pbs_end_time`, etc.

        """
        assert self._attached, ("Task.update_state() called on detached task %s." % self)
        assert hasattr(self._controller, 'update_job_state'), \
               ("Invalid `_controller` object '%s' in Task %s" % (self._controller, self))
        self._controller.update_job_state(self, **extra_args)


    def kill(self, **extra_args):
        """
        Terminate the computational job associated with this task.

        See :meth:`gc3libs.Core.kill` for a full explanation.
        """
        assert self._attached, ("Task.kill() called on detached task %s." % self)
        assert hasattr(self._controller, 'kill'), \
               ("Invalid `_controller` object '%s' in Task %s" % (self._controller, self))
        self._controller.kill(self, **extra_args)


    def fetch_output(self, output_dir=None, overwrite=False, **extra_args):
        """
        Retrieve the outputs of the computational job associated with
        this task into directory `output_dir`, or, if that is `None`,
        into the directory whose path is stored in instance attribute
        `.output_dir`.

        If the execution state is `TERMINATING`, transition the state to
        `TERMINATED` (which runs the appropriate hook).

        See :meth:`gc3libs.Core.fetch_output` for a full explanation.

        :return: Path to the directory where the job output has been
                 collected.
        """
        #result = self._controller.fetch_output(self, output_dir, overwrite, **extra_args)
        if self.execution.state == Run.State.TERMINATED:
            return self.output_dir
        # advance state to TERMINATED
        if self.execution.state == Run.State.TERMINATING:
            self.output_dir = self._get_download_dir(output_dir)
            self.execution.info = ("Final output downloaded to '%s'" % self.output_dir)
            self.execution.state = Run.State.TERMINATED
            self.changed = True
            return self.output_dir
        else:
            download_dir = self._get_download_dir(output_dir)
            self.execution.info = ("Output snapshot downloaded to '%s'" % download_dir)
            return download_dir

    def _get_download_dir(self, download_dir):
        """
        Return a directory path where to download this Task's output files.

        If the given `download_dir` is not None, return that.  Otherwise,
        return the directory saved on this object in attribute `output_dir`.
        If all else fails, raise `gc3libs.exceptions.InvalidArgument`.
        """
        # determine download dir
        if download_dir is not None:
            return download_dir
        else:
            try:
                return self.output_dir
            except AttributeError:
                raise gc3libs.exceptions.InvalidArgument(
                    "`Task._get_download_dir()` called with no explicit download directory,"
                    " but object '%s' has no `output_dir` attribute set either."
                    % (self, type(self)))


    def peek(self, what='stdout', offset=0, size=None, **extra_args):
        """
        Download `size` bytes (at offset `offset` from the start) from
        the associated job standard output or error stream, and write them
        into a local file.  Return a file-like object from which the
        downloaded contents can be read.

        See :meth:`gc3libs.Core.peek` for a full explanation.
        """
        assert self._attached, ("Task.peek() called on detached task %s." % self)
        assert hasattr(self._controller, 'peek'), \
               ("Invalid `_controller` object '%s' in Task %s" % (self._controller, self))
        return self._controller.peek(self, what, offset, size, **extra_args)


    def free(self, **extra_args):
        """
        Release any remote resources associated with this task.

        See :meth:`gc3libs.Core.free` for a full explanation.
        """
        return


    # convenience methods, do not really add any functionality over
    # what's above

    def progress(self):
        """
        Advance the associated job through all states of a regular
        lifecycle. In detail:

          1. If `execution.state` is `NEW`, the associated job is started.
          2. The state is updated until it reaches `TERMINATED`
          3. Output is collected and the final returncode is returned.

        An exception `TaskError` is raised if the job hits state
        `STOPPED` or `UNKNOWN` during an update in phase 2.

        When the job reaches `TERMINATING` state, the output is
        retrieved; if this operation is successfull, state is advanced
        to `TERMINATED`.

        oNCE the job reaches `TERMINATED` state, the return code
        (stored also in `.returncode`) is returned; if the job is not
        yet in `TERMINATED` state, calling `progress` returns `None`.

        :raises: exception :class:`UnexpectedStateError` if the
                 associated job goes into state `STOPPED` or `UNKNOWN`

        :return: final returncode, or `None` if the execution
                 state is not `TERMINATED`.

        """
        # first update state, we'll submit NEW jobs last, so that the
        # state is not updated immediately after submission as ARC
        # does not cope well with this...
        if self.execution.state in [ Run.State.SUBMITTED,
                                     Run.State.RUNNING,
                                     Run.State.STOPPED,
                                     Run.State.UNKNOWN ]:
            self.update_state()
        # now "do the right thing" based on actual state
        if self.execution.state in [ Run.State.STOPPED,
                                     Run.State.UNKNOWN ]:
            raise gc3libs.exceptions.UnexpectedStateError(
                "Task '%s' entered `%s` state." % (self, self.execution.state))
        elif self.execution.state == Run.State.NEW:
            self.submit()
        elif self.execution.state == Run.State.TERMINATING:
            self.fetch_output()
            return self.execution.returncode


    def wait(self, interval=60):
        """
        Block until the associated job has reached `TERMINATED` state,
        then return the job's return code.  Note that this does not
        automatically fetch the output.

        :param integer interval: Poll job state every this number of seconds
        """
        # FIXME: I'm not sure how to deal with this... Ideally, this
        # call should suspend the current thread and wait for
        # notifications from the Engine, but:
        #  - there's no way to tell if we are running threaded,
        #  - `self._controller` could be a `Core` instance, thus not capable
        #    of running independently.
        # For now this is a poll+sleep loop, but we certainly need to revise it.
        while True:
            self.update_state()
            if self.execution.state == Run.State.TERMINATED:
                return self.returncode
            time.sleep(interval)


    # State transition handlers.
    #

    def new(self):
        """
        Called when the job state is (re)set to `NEW`.

        Note this will not be called when the application object is
        created, rather if the state is reset to `NEW` after it has
        already been submitted.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass

    def submitted(self):
        """
        Called when the job state transitions to `SUBMITTED`, i.e.,
        the job has been successfully sent to a (possibly) remote
        execution resource and is now waiting to be scheduled.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass


    def running(self):
        """
        Called when the job state transitions to `RUNNING`, i.e., the
        job has been successfully started on a (possibly) remote
        resource.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass


    def stopped(self):
        """
        Called when the job state transitions to `STOPPED`, i.e., the
        job has been remotely suspended for an unknown reason and
        cannot automatically resume execution.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass


    def terminating(self):
        """
        Called when the job state transitions to `TERMINATING`, i.e.,
        the remote job has finished execution (with whatever exit
        status, see `returncode`) but output has not yet been
        retrieved.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass


    def terminated(self):
        """
        Called when the job state transitions to `TERMINATED`, i.e.,
        the job has finished execution (with whatever exit status, see
        `returncode`) and the final output has been retrieved.

        The location where the final output has been stored is
        available in attribute `self.output_dir`.

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass

    def unknown(self):
        """
        Called when the job state transitions to `UNKNOWN`, i.e.,
        the job has not been updated for a certain period of time
        thus it is placed in UNKNOWN state.

        Two possible ways of changing from this state:
        1) next update cycle, job status is updated from the remote
        server
        2) derive this method for Application specific logic to deal
        with this case

        The default implementation does nothing, override in derived
        classes to implement additional behavior.
        """
        pass

def configure_logger(level=logging.ERROR,
                     name=None,
                     format=(os.path.basename(sys.argv[0])
                             + ': [%(asctime)s] %(levelname)-8s: %(message)s'),
                     datefmt='%Y-%m-%d %H:%M:%S'):
    """
    Configure the ``gc3.gc3libs`` logger.

    Arguments `level`, `format` and `datefmt` set the corresponding
    arguments in the `logging.basicConfig()` call.

    If a user configuration file exists in file NAME.log.conf in the
    ``Default.RCDIR`` directory (usually ``~/.gc3``), it is read and
    used for more advanced configuration; if it does not exist, then a
    sample one is created.
    """
    if name is None:
        name = os.path.basename(sys.argv[0])
    log_conf = os.path.join(Default.RCDIR, name + '.log.conf')
    logging.basicConfig(level=level, format=format, datefmt=datefmt)
    deploy_configuration_file(log_conf, "logging.conf.example")
    logging.config.fileConfig(log_conf, {
        'RCDIR': Default.RCDIR,
        'HOMEDIR': os.path.expandvars('$HOME'),
        })
    log = logging.getLogger("gc3.gc3libs")
    log.setLevel(level)
    log.propagate = 1
    # Up to Python 2.5, the `logging` library disables all existing
    # loggers upon reconfiguration, and fails to re-create them when
    # getLogger() is called again.  We work around this the hard way:
    # using an undocumented internal variable, ignore errors, and hope
    # for the best.
    try:
        log.disabled = 0
    except:
        pass
    # due to a bug in Python 2.4.x (see
    # https://bugzilla.redhat.com/show_bug.cgi?id=573782 )
    # we need to disable `logging` reporting of exceptions.
    try:
        version_info = sys.version_info
    except AttributeError:
        version_info = (1, 5) # 1.5 or earlier
    if version_info < (2, 5):
        logging.raiseExceptions = False


# when used in the `output` attribute of an application,
# it stands for "fetch the whole contents of the remote
# directory"
ANY_OUTPUT = '*'

class Application(Task):
    """
    Support for running a generic application with the GC3Libs.
    The following parameters are *required* to create an `Application`
    instance:

    `arguments`
      List or sequence of program arguments. The program to execute is
      the first one.; any object in the list will be converted to
      string via Python's `str()`.

    `inputs`
      Files that will be copied to the remote execution node before
      execution starts.

      There are two possible ways of specifying the `inputs` parameter:

      * It can be a Python dictionary: keys are local file paths or
        URLs, values are remote file names.

      * It can be a Python list: each item in the list should be a
        pair `(source, remote_file_name)`: the `source` can be a
        local file or a URL; `remote_file_name` is the path
        (relative to the execution directory) where `source` will be
        downloaded.  If `remote_file_name` is an absolute path, an
        :class:`InvalidArgument` error is raised.

        A single string `file_name` is allowed instead of the pair
        and results in the local file `file_name` being copied to
        `file_name` on the remote host.

    `outputs`
      Files and directories that will be copied from the remote
      execution node back to the local computer (or a
      network-accessible server) after execution has completed.
      Directories are copied recursively.

      There are three possible ways of specifying the `outputs` parameter:

      * It can be a Python dictionary: keys are remote file or
        directory paths (relative to the execution directory), values
        are corresponding local names.

      * It can be a Python list: each item in the list should be a
        pair `(remote_file_name, destination)`: the `destination`
        can be a local file or a URL; `remote_file_name` is the path
        (relative to the execution directory) that will be uploaded
        to `destination`.  If `remote_file_name` is an absolute
        path, an :class:`InvalidArgument` error is raised.

        A single string `file_name` is allowed instead of the pair
        and results in the remote file `file_name` being copied to
        `file_name` on the local host.

      * The constant `gc3libs.ANY_OUTPUT` which instructs GC3Libs to
        copy every file in the remote execution directory back to the
        local output path (as specified by the `output_dir` attribute).

      Note that no errors will be raised if an output file is not present.
      Override the `terminated`:meth: method to raise errors for reacting
      on this kind of failures.

    `output_dir`
      Path to the base directory where output files will be downloaded.
      Output file names are interpreted relative to this base directory.

    `requested_cores`,`requested_memory`,`requested_walltime`
      specify resource requirements for the application:
      * the number of independent execution units (CPU cores),
      * amount of memory (as a `gc3libs.quantity.Memory`:class: object),
      * amount of wall-clock time to allocate for the computational job (as a `gc3libs.quantity.Duration`:class: object).

    The following optional parameters may be additionally
    specified as keyword arguments and will be given special
    treatment by the `Application` class logic:

    `requested_architecture`
      specify that this application can only be executed on a certain
      processor architecture; see `Run.Arch`:class: for a list of
      possible values.  The default value `None` means that any
      architecture is valid, i.e., there are no requirements on the
      processor architecture.

    `environment`
      a list of pairs `(name, value)`: the
      environment variable whose name is given by the contents of
      the string `name` will be defined as the content of string
      `value` (i.e., as if "export name=value" was executed prior
      to starting the application).  Alternately, one can pass in
      a list of strings of the form "name=value".

    `output_base_url`
      if not `None`, this is prefixed to all output files (except
      stdout and stderr, which are always retrieved), so, for instance,
      having `output_base_url="gsiftp://example.org/data"` will upload
      output files into that remote directory.

    `stdin`
      file name of a file whose contents will be fed as
      standard input stream to the remote-executing process.

    `stdout`
      name of a file where the standard output stream of
      the remote executing process will be redirected to; will be
      automatically added to `outputs`.

    `stderr`
      name of a file where the standard error stream of
      the remote executing process will be redirected to; will be
      automatically added to `outputs`.

    `join`
      if this evaluates to `True`, then standard error is
      redirected to the file specified by `stdout` and `stderr` is
      ignored.  (`join` has no effect if `stdout` is not given.)

    `tags`
      list of tag names (string) that must be present on a
      resource in order to be eligible for submission; in the ARC
      backend, tags are interpreted as run-time environments (RTE) to
      request.

    Any other keyword arguments will be set as instance attributes,
    but otherwise ignored by the `Application` constructor.

    After successful construction, an `Application` object is
    guaranteed to have the following instance attributes:

    `arguments`
      list of strings specifying command-line arguments for executable
      invocation. The first element must be the executable.

    `inputs`
      dictionary mapping source URL (a `gc3libs.url.Url`:class:
      object) to a remote file name (a string); remote file names are
      relative paths (root directory is the remote job folder)

    `outputs`
      dictionary mapping remote file name (a string) to a destination
      (a `gc3libs.url.Url`:class:); remote file names are relative
      paths (root directory is the remote job folder)

    `output_dir`
      Path to the base directory where output files will be
      downloaded.  Output file names (those which are not URLs) are
      interpreted relative to this base directory.

    `execution`
      a `Run` instance; its state attribute is initially set to ``NEW``
      (Actually inherited from the `Task`:class:)

     `environment`
      dictionary mapping environment variable names to the requested
      value (string); possibly empty

    `stdin`
      `None` or a string specifying a (local) file name.  If `stdin`
      is not None, then it matches a key name in `inputs`

    `stdout`
      `None` or a string specifying a (remote) file name.  If `stdout`
      is not None, then it matches a key name in `outputs`

    `stderr`
      `None` or a string specifying a (remote) file name.  If `stdout`
      is not None, then it matches a key name in `outputs`

    `join`
      boolean value, indicating whether `stdout` and `stderr` are
      collected into the same file

    `tags`
      list of strings specifying the tags to request in each resource
      for submission; possibly empty.
    """

    application_name = 'generic'
    """
    A name for applications of this class.

    This string is used as a prefix for configuration items related to
    this application in configured resources.  For example, if the
    `application_name` is ``foo``, then the application interface code
    in GC3Pie might search for ``foo_cmd``, ``foo_extra_args``, etc.
    See `qsub_sge`:meth: for an actual example.
    """

    def __init__(self, arguments, inputs, outputs, output_dir, **extra_args):
        # required parameters
        if isinstance(arguments, types.StringTypes):
            arguments = shlex.split(arguments)

        if 'executable' in extra_args:
            gc3libs.log.warning(
                "The `executable` argument is not supported anymore in GC3Pie 2.0."
                " Please adapt your code and use `arguments` only.")
            arguments = [ extra_args['executable'] ] + list(arguments)

        self.arguments = [ str(x) for x in arguments ]

        self.inputs = Application._io_spec_to_dict(UrlKeyDict, inputs, True)
        self.outputs = Application._io_spec_to_dict(UrlValueDict, outputs, False)

        # check that remote entries are all distinct
        # (can happen that two local paths are mapped to the same remote one)
        if len(self.inputs.values()) != len(set(self.inputs.values())):
            # try to build an exact error message
            inv = { }
            for l, r in self.inputs.iteritems():
                if r in inv:
                    raise DuplicateEntryError("Local inputs '%s' and '%s'"
                                              " map to the same remote path '%s'"
                                              % (l, inv[r], r))
                else:
                    inv[r] = l

        # ensure remote paths are not absolute
        for r_path in self.inputs.itervalues():
            if os.path.isabs(r_path):
                raise gc3libs.exceptions.InvalidArgument(
                    "Remote paths not allowed to be absolute: %s" % r_path)

        # check that local entries are all distinct
        # (can happen that two remote paths are mapped to the same local one)
        if len(self.outputs.values()) != len(set(self.outputs.values())):
            # try to build an exact error message
            inv = { }
            for r, l in self.outputs.iteritems():
                if l in inv:
                    raise DuplicateEntryError("Remote outputs '%s' and '%s'"
                                              " map to the same local path '%s'"
                                              % (r, inv[l], l))
                else:
                    inv[l] = r

        # ensure remote paths are not absolute
        for r_path in self.outputs.iterkeys():
            if os.path.isabs(r_path):
                raise gc3libs.exceptions.InvalidArgument(
                    "Remote paths not allowed to be absolute")

        self.output_dir = output_dir

        # optional params
        self.output_base_url = extra_args.pop('output_base_url', None)

        self.requested_cores = int(extra_args.pop('requested_cores', 1))
        self.requested_memory = extra_args.pop('requested_memory', None)
        assert (self.requested_memory is None
                or isinstance(self.requested_memory, gc3libs.quantity.Memory)), \
            ("Expected `Memory` instance for `requested_memory, got %r %s instead."
             % (self.requested_memory, type(self.requested_memory)))
        self.requested_walltime = extra_args.pop('requested_walltime', None)
        assert (self.requested_walltime is None
                or isinstance(self.requested_walltime, gc3libs.quantity.Duration)), \
            ("Expected `Duration` instance for `requested_walltime, got %r %s instead."
             % (self.requested_memory, type(self.requested_memory)))
        self.requested_architecture = extra_args.pop('requested_architecture', None)
        if self.requested_architecture is not None \
               and self.requested_architecture not in [ Run.Arch.X86_32, Run.Arch.X86_64 ]:
            raise gc3libs.exceptions.InvalidArgument(
                "Architecture must be either '%s' or '%s'"
                % (Run.Arch.X86_32, Run.Arch.X86_64))

        self.environment = extra_args.pop('environment', dict())
        self.environment = dict(Application._to_env_pair(x)
                                for x in self.environment.items())

        self.join = extra_args.pop('join', False)
        self.stdin = extra_args.pop('stdin', None)
        if self.stdin and (self.stdin not in self.inputs):
            self.inputs[self.stdin] = os.path.basename(self.stdin)
        self.stdout = extra_args.pop('stdout', None)
        if self.stdout is not None and os.path.isabs(self.stdout):
            raise InvalidArgument(
                "Absolute path '%s' passed as `Application.stdout`"
                % self.stdout)
        if ((self.stdout is not None)
            and (gc3libs.ANY_OUTPUT not in self.outputs)
            and (self.stdout not in self.outputs)):
            self.outputs[self.stdout] = self.stdout

        self.stderr = extra_args.pop('stderr', None)
        if self.stderr == self.stdout or self.stderr == subprocess.STDOUT:
            self.join = True
            self.stderr = self.stdout

        if self.stderr is not None and os.path.isabs(self.stderr):
            raise InvalidArgument(
                "Absolute path '%s' passed as `Application.stderr`"
                % self.stderr)
        if ((self.stderr is not None)
            and (gc3libs.ANY_OUTPUT not in self.outputs)
            and (self.stderr not in self.outputs)):
            self.outputs[self.stderr] = self.stderr

        self.tags = extra_args.pop('tags', list())

        if 'jobname' in extra_args:
            jobname = extra_args['jobname']
            # Check whether the first character of a jobname is an
            # integer. SGE does not allow job names to start with a
            # number, so add a prefix...
            if len(jobname) == 0:
                gc3libs.log.warning("Empty string passed as jobname to %s", self)
                jobname = "GC3Pie.%s.%s" % (self.__class__.__name__, id(self))
            elif str(jobname)[0] in string.digits:
                jobname = "GC3Pie.%s" % jobname
            extra_args['jobname'] = jobname

        # task setup; creates the `.execution` attribute as well
        Task.__init__(self, **extra_args)

        # for k,v in self.outputs.iteritems():
        #     gc3libs.log.debug("outputs[%s]=%s", repr(k), repr(v))
        # for k,v in self.inputs.iteritems():
        #     gc3libs.log.debug("inputs[%s]=%s", repr(k), repr(v))


    @staticmethod
    def _to_env_pair(val):
        if isinstance(val, tuple):
            return val
        else:
            # assume `val` is a string
            return tuple(val.split('=', 1))


    @staticmethod
    def _io_spec_to_dict(ctor, spec, force_abs):
        """
        (This function is only used for internal processing of `input`
        and `output` fields.)

        Return a dictionary formed by pairs `URL:name` or `name:URL`.
        The `URL` part is a tuple as returned by functions `urlparse`
        and `urlsplit` in the Python standard module
        `urlparse`:module: -- `name` is a string that should be
        interpreted as a filename (relative to the job execution
        directory).

        Argument `ctor` is the constructor for the dictionary class to
        return; `gc3libs.url.UrlKeyDict` and
        `gc3libs.url.UrlValueDict` are valid values here.

        Argument `spec` is either a list or a Python `dict` instance.

        If a Python `dict` is given, then it is copied into an
        `gc3libs.url.UrlDict`, and that copy is returned::

          >>> d1 = { '/tmp/1':'1', '/tmp/2':'2' }
          >>> d2 = Application._io_spec_to_dict(gc3libs.url.UrlKeyDict, d1, True)
          >>> isinstance(d2, gc3libs.url.UrlKeyDict)
          True
          >>> for key in sorted(d2.keys()): print key.path
          /tmp/1
          /tmp/2

        If `spec` is a list, each element can either be a tuple
        `(path, name)`, or a string `path`, which is converted to a
        tuple `(path, name)` by setting `name =
        os.path.basename(path)`::

          >>> l1 = [ ('/tmp/1', '1'), '/tmp/2' ]
          >>> d3 = Application._io_spec_to_dict(gc3libs.url.UrlKeyDict, l1, True)
          >>> d3 == d2
          True

        If `force_abs` is `True`, then all paths are converted to
        absolute ones in the dictionary keys; otherwise they are
        stored unchanged.
        """
        try:
            # is `spec` dict-like?
            return ctor(((str(k), str(v)) for k,v in spec.iteritems()),
                        force_abs=force_abs)
        except UnicodeError, err:
            raise gc3libs.exceptions.InvalidValue(
                "Use of non-ASCII file names is not (yet) supported in GC3Pie: %s: %s"
                % (err.__class__.__name__, str(err)))
        except AttributeError:
            # `spec` is a list-like
            def convert_to_tuple(val):
                if isinstance(val, types.StringTypes):
                    l = str(val)
                    r = os.path.basename(l)
                    return (l, r)
                else:
                    return (str(val[0]), str(val[1]))
            return ctor((convert_to_tuple(x) for x in spec),
                        force_abs=force_abs)


    def __str__(self):
        try:
            return str(self.persistent_id)
        except AttributeError:
            return safe_repr(self)


    def compatible_resources(self, resources):
        """
        Return a list of compatible resources.
        """
        selected = [ ]
        for lrms in resources:
            assert (lrms is not None), \
                "Application.compatible_resources():" \
                " expected `LRMS` object, got `None` instead."
            gc3libs.log.debug(
                "Checking resource '%s' for compatibility with application requirements"
                % lrms.name)
            # if architecture is specified, check that it matches the resource one
            if (self.requested_architecture is not None
                and self.requested_architecture not in lrms.architecture):
                gc3libs.log.info("Rejecting resource '%s': requested a different architecture (%s) than what resource provides (%s)"
                                 % (lrms.name, self.requested_architecture,
                                    str.join(', ', [str(arch) for arch in lrms.architecture ])))
                continue
            # check that Application requirements are within resource limits
            if self.requested_cores > lrms.max_cores_per_job:
                gc3libs.log.info("Rejecting resource '%s': requested more cores (%d) that resource provides (%d)"
                                 % (lrms.name, self.requested_cores, lrms.max_cores_per_job))
                continue
            if (self.requested_memory is not None
                and self.requested_memory > self.requested_cores * lrms.max_memory_per_core):
                gc3libs.log.info("Rejecting resource '%s': requested more memory (%s) that resource provides (%s, %s per CPU core)"
                                 % (lrms.name, self.requested_memory, self.requested_cores*lrms.max_memory_per_core, lrms.max_memory_per_core))
                continue
            if (self.requested_walltime is not None
                and self.requested_walltime > lrms.max_walltime):
                gc3libs.log.info("Rejecting resource '%s': requested a longer duration (%s) that resource provides (%s)"
                                 % (lrms.name, self.requested_walltime, lrms.max_walltime))
                continue
            if not lrms.validate_data(self.inputs.keys()) or not lrms.validate_data(self.outputs.values()):
                gc3libs.log.info("Rejecting resource '%s': input/output data protocol not supported."
                                 % lrms.name)
                continue

            selected.append(lrms)

        return selected


    @staticmethod
    def _cmp_resources(a,b):
        """
        Compare resources `a` and `b` and return -1,0,1 accordingly
        (see doc for the Python standard function `cmp`).

        Computational resource `a` is preferred over `b` if it has less
        queued jobs from the same user; failing that, if it has more free
        slots; failing that, if it has less queued jobs (in total);
        finally, should all preceding parameters compare equal, `a` is
        preferred over `b` if it has less running jobs from the same user.
        """
        a_ = (a.user_queued, -a.free_slots, a.queued, a.user_run)
        b_ = (b.user_queued, -b.free_slots, b.queued, b.user_run)
        return cmp(a_, b_)

    def rank_resources(self, resources):
        """
        Sort the given resources in order of preference.

        By default, less-loaded resources come first;
        see `_cmp_resources`.
        """
        # shift lrms that are already in application.execution_targets
        # to the bottom of the list
        selected = sorted(resources, cmp=self._cmp_resources)
        if '_execution_targets' in self.execution:
            for lrms in selected:
                if (hasattr(lrms, 'frontend')
                     and lrms.frontend in self.execution._execution_targets):
                    # append resource to the bottom of the list
                    selected.remove(lrms)
                    selected.append(lrms)
        return selected

    def fetch_output(self, download_dir, overwrite, **extra_args):
        """
        Calls the corresponding method of the controller.
        """
        return self._controller.fetch_output(self, download_dir, overwrite, **extra_args)

    ##
    ## backend interface methods
    ##

    def xrsl(self, resource):
        """
        Return a string containing an xRSL sequence, suitable for
        submitting an instance of this application through ARC's
        ``ngsub`` command.

        The default implementation produces XRSL content based on
        the construction parameters; you should override this method
        to produce XRSL tailored to your application.

        .. warning::

          WARNING: ARClib SWIG bindings cannot resolve the overloaded
          constructor if the xRSL string argument is a Python 'unicode'
          object; if you overload this method, force the result to be
          a 'str'!
        """
        # build `xrsl` incrementally; concatenate all parts into a
        # single string at the end
        xrsl= [
            '&',
            '(executable="/bin/sh")',
            '(gmlog=".gc3pie_arc")', # XXX: should check if conflicts with any input/output files
            ]
        xrsl.append('(arguments="-c" "%s")' % str.join(' ', self.arguments))
        # preserve execute permission on all input files
        executables = []
        if self.has_key('executables'):
            # there are already references to files that should be
            # executables
            # e.g. a reference to a remote file that cannot be checked locally
            executables = self.executables
        for l, r in self.inputs.iteritems():
            if os.access(l.path, os.X_OK):
                executables.append(r)
        if len(executables) > 0:
            xrsl.append('(executables=%s)'
                        % str.join(' ', [('"%s"' % x) for x in executables]))
        if self.stdin:
            xrsl.append('(stdin="%s")' % self.stdin)
        # XXX: this can go away when we have the ternary operator
        # `x = a if y else b` (Python 2.5)
        if self.join:
            xrsl.append('(join="yes")')
        else:
            xrsl.append('(join="no")')
        if self.stdout:
            xrsl.append('(stdout="%s")' % self.stdout)
        else:
            xrsl.append('(stdout="/dev/null")')
        if self.stderr and not self.join:
            xrsl.append('(stderr="%s")' % self.stderr)
        if len(self.inputs) > 0:
            xrsl.append('(inputFiles=%s)'
                        % str.join(' ', [ ('("%s" "%s")' % (r, l))
                                          for (l,r) in self.inputs.items() ]))
        if len(self.outputs) > 0:
            # XXX: this can go away when we have the ternary operator
            # `x = a if y else b` (Python 2.5)
            if self.output_base_url is None:
                def output_url(l, r):
                    if l.scheme == 'file':
                        return ''
                    else:
                        return l
            else:
                def output_url(l, r):
                    if l.scheme == 'file':
                        return os.path.join(self.output_base_url,
                                            utils.ifelse(l, l, r))
                    else:
                        return l

            def relpath(r):
                if r == gc3libs.ANY_OUTPUT:
                    return '/'
                else:
                    return r
            # filter out stdout/stderr (they are automatically
            # retrieved) and then check again
            outputs_ = [ ('("%s" "%s")' % (relpath(r), output_url(l, r)))
                         for (r,l) in [ (remotename, localname)
                                        for remotename,localname
                                        in self.outputs.iteritems()
                                        if (remotename != self.stdout
                                            and remotename != self.stderr)]]
            if len(outputs_) > 0:
                xrsl.append('(outputFiles=%s)' % str.join(' ', outputs_))
        if len(self.tags) > 0:
            xrsl += [('(runTimeEnvironment="%s")' % rte) for rte in self.tags ]
        if len(self.environment) > 0:
            xrsl.append('(environment=%s)' %
                        str.join(' ', [ ('("%s" "%s")' % kv) for kv in self.environment.iteritems() ]))
        if self.requested_walltime:
            # xRSL assumes minutes by default
            xrsl.append('(wallTime="%d")' % self.requested_walltime.amount(minutes))
        if self.requested_memory:
            # ARC's "memory" is memory per "rank" (= core in MPI-speak)
            xrsl.append('(memory="%d")'
                        % (self.requested_memory.amount(MB) / self.requested_cores))
        if self.requested_cores:
            xrsl.append('(count="%d")' % self.requested_cores)
        # XXX: the xRSL specification states that the "architecture" value
        # is matched against the value reported as `uname -a` on the cluster,
        # but different Linux distributions use "i386", "i586" and "i686"
        # as `uname -a` values, so there is no single value that can
        # match any x86 arch...
        if self.requested_architecture is not None:
            xrsl.append('(architecture="%s")' % self.requested_architecture)
        if 'jobname' in self:
            xrsl.append('(jobname="%s")' % self.jobname)

        # XXX: experimental
        # this should be harmless if cache registration would not work
        xrsl.append('(cache="yes")')

        # WARNING: ARClib SWIG bindings cannot resolve the overloaded
        # constructor if the argument is a Python "unicode" object;
        # force it to be a "str"!
        return str.join(' ', xrsl)


    def cmdline(self, resource):
        """
        Return list of command-line arguments for invoking the application.

        This is exactly the *argv*-vector of the application process:
        the application command name is included as first item (index
        0) of the list, further items are command-line arguments.

        Hence, to get a UNIX shell command-line, just concatenate the
        elements of the list, separating them with spaces.

        """
        return self.arguments[:]


    def qsub_sge(self, resource, **extra_args):
        """
        Get an SGE ``qsub`` command-line invocation for submitting an
        instance of this application.

        Return a pair `(cmd_argv, app_argv)`.  Both `cmd_argv` and
        `app_argv` are *argv*-lists: the command name is included as
        first item (index 0) of the list, further items are
        command-line arguments; `cmd_argv` is the *argv*-list for the
        submission command (excluding the actual application command
        part); `app_argv` is the *argv*-list for invoking the
        application.  By overriding this method, one can add futher
        resource-specific options at the end of the `cmd_argv`
        *argv*-list.

        In the construction of the command-line invocation, one should
        assume that all the input files (as named in `Application.inputs`)
        have been copied to the current working directory, and that output
        files should be created in this same directory.

        The default implementation just prefixes any output from the
        `cmdline` method with an SGE ``qsub`` invocation of the form
        ``qsub -cwd -S /bin/sh`` + resource limits.  Note that
        *there is no generic way of requesting a certain number of cores*
        in SGE: it all depends on the installed parallel environment, and
        these are totally under control of the local sysadmin;
        therefore, any request for cores is ignored and a warning is
        logged.

        Override this method in application-specific classes to
        provide appropriate invocation templates and/or add different
        submission options.
        """
        qsub = list(resource.qsub)
        qsub += ['-cwd', '-S', '/bin/sh']
        if self.requested_walltime:
            # SGE uses `s_rt` for wall-clock time limit, expressed in seconds
            qsub += ['-l', 's_rt=%d' % self.requested_walltime.amount(seconds)]
        if self.requested_memory:
            # SGE uses `mem_free` for memory limits; 'M' suffix allowed for Megabytes
            # XXX: there are a number of problems here:
            #   - `mem_free` might not be requestable, i.e., submission will fail
            #   - `mem_free` might be a JOB consumable, meaning the value should be the total amount of memory requested by the job
            #   - in the end it's all matter of local configuration, so we might need to request `h_vmem` (job total) *and* `virtual_free` (per-slot consumable) ...
            # Let's make whatever works in our cluster, and see how we can
            # extend/change it when issue reports come...
            qsub += ['-l', ('mem_free=%dM'
                            % (self.requested_memory.amount(MB) / self.requested_cores))]
        if self.join:
            qsub += ['-j', 'yes']
        if self.stdout:
            qsub += ['-o', '%s' % self.stdout]
        if self.stdin:
            # `self.stdin` is the full pathname on the GC3Pie client host;
            # it is copied to its basename on the execution host
            qsub += ['-i', '%s' % os.path.basename(self.stdin)]
        if self.stderr:
            # from the qsub(1) man page: "If both the -j y and the -e
            # options are present, Grid Engine sets but ignores the
            # error-path attribute."
            qsub += ['-e', '%s' % self.stderr]
        if self.requested_cores != 1:
            pe_cfg_name = (self.application_name + '_pe')
            if pe_cfg_name in resource:
                pe_name = resource.get(pe_cfg_name)
            else:
                pe_name = resource.get('default_pe')
                if pe_name is not None:
                    # XXX: overly verbose reporting?
                    log.info(
                        "Application %s requested %d cores,"
                        " but no '%s' configuration item is defined on resource '%s';"
                        " using the 'default_pe' setting to submit the parallel job.",
                        self, self.requested_cores, pe_cfg_name, resource.name)
                else:
                    raise gc3libs.exceptions.InternalError(
                        "Application %s requested %d cores,"
                        " but neither '%s' nor 'default_pe' appear in the configuration"
                        " of resource '%s'.  Please fix the configuration and retry."
                        % (self, self.requested_cores, pe_cfg_name, resource.name))
            qsub += ['-pe', pe_name, ('%d' % self.requested_cores)]
        if 'jobname' in self and self.jobname:
            qsub += ['-N', '%s' % self.jobname]
        return (qsub, self.cmdline(resource))


    def bsub(self, resource, _suppress_warning=False, **extra_args):
        # XXX: the `_suppress_warning` switch is only provided for
        # some applications to make use of this generic method without
        # logging the user-level warning, because, e.g., it has already
        # been taken care in some other way (cf. GAMESS' `qgms`).
        # Use with care and don't depend on it!
        """
        Get an LSF ``qsub`` command-line invocation for submitting an
        instance of this application.  Return a pair `(cmd_argv,
        app_argv)`, where `cmd_argv` is a list containing the
        *argv*-vector of the command to run to submit an instance of
        this application to the LSF batch system, and `app_argv` is
        the *argv*-vector to use when invoking the application.

        In the construction of the command-line invocation, one should
        assume that all the input files (as named in `Application.inputs`)
        have been copied to the current working directory, and that output
        files should be created in this same directory.

        The default implementation just prefixes any output from the
        `cmdline` method with an LSF ``bsub`` invocation of the form
        ``bsub -cwd . -L /bin/sh`` + resource limits.

        Override this method in application-specific classes to
        provide appropriate invocation templates and/or add
        resource-specific submission options.
        """
        bsub = list(resource.bsub)
        bsub += ['-cwd', '.', '-L', '/bin/sh', '-n', ('%d' % self.requested_cores)]
        if self.requested_walltime:
            # LSF wants walltime as HH:MM (days expressed as many hours)
            hs = int(self.requested_walltime.amount(hours))
            ms = int(self.requested_walltime.amount(minutes)) % 60
            bsub += ['-W', ('%02d:%02d' % (hs, ms))]
        if self.requested_memory:
            # LSF uses `rusage[mem=...]` for memory limits (number of MBs)
            bsub += ['-R', ('rusage[mem=%d]' % self.requested_memory.amount(MB))]
        if self.stdout:
            bsub += ['-oo', ('%s' % self.stdout)]
            if not self.join and not self.stderr:
                # LSF joins STDERR and STDOUT by default, so redirect STDERR away
                bsub += ['-eo', '/dev/null']
        if self.stdin:
            # `self.stdin` is the full pathname on the GC3Pie client host;
            # it is copied to its basename on the execution host
            bsub += ['-i', ('%s' % os.path.basename(self.stdin))]
        if self.stderr:
            bsub += ['-eo', ('%s' % self.stderr)]
        if 'jobname' in self and self.jobname:
            bsub += ['-J', ('%s' % self.jobname)]
        return (bsub, self.cmdline(resource))


    def qsub_pbs(self, resource, _suppress_warning=False, **extra_args):
        """
        Similar to `qsub_sge()`, but for the PBS/TORQUE resource manager.
        """
        qsub = list(resource.qsub)
        if self.requested_walltime:
            qsub += ['-l', 'walltime=%s' % (self.requested_walltime.amount(seconds))]
        if self.requested_memory:
            qsub += ['-l', 'mem=%dmb' % self.requested_memory.amount(MB)]
        if self.stdin:
            # `self.stdin` is the full pathname on the GC3Pie client host;
            # it is copied to its basename on the execution host
            # XXX: this is wrong, as it redirects STDIN of the `qsub` process!
            #qsub += ['<', '%s' % os.path.basename(self.stdin)]
            raise NotImplementedError("STDIN redirection is currently not handled by the PBS/TORQUE backend!")
        if self.stderr:
            # from the qsub(1) man page: "If both the -j y and the -e
            # options are present, Grid Engine sets but ignores the
            # error-path attribute."
            qsub += ['-e', '%s' % self.stderr]
        if self.stdout:
            # from the qsub(1) man page: "If both the -j y and the -e
            # options are present, Grid Engine sets but ignores the
            # error-path attribute."
            qsub += ['-o', '%s' % self.stdout]
        if self.requested_cores > 1:
            qsub += ['-l', 'nodes=%d' % self.requested_cores]
        if 'jobname' in self and self.jobname:
            qsub += ['-N', '"%s"' % self.jobname]
        return (qsub, self.cmdline(resource))


    def sbatch(self, resource, **extra_args):
        """
        Get a SLURM ``sbatch`` command-line invocation for submitting an
        instance of this application.

        Return a pair `(cmd_argv, app_argv)`.  Both `cmd_argv` and
        `app_argv` are *argv*-lists: the command name is included as
        first item (index 0) of the list, further items are
        command-line arguments; `cmd_argv` is the *argv*-list for the
        submission command (excluding the actual application command
        part); `app_argv` is the *argv*-list for invoking the
        application.  By overriding this method, one can add futher
        resource-specific options at the end of the `cmd_argv`
        *argv*-list.

        In the construction of the command-line invocation, one should
        assume that all the input files (as named in `Application.inputs`)
        have been copied to the current working directory, and that output
        files should be created in this same directory.

        Override this method in application-specific classes to
        provide appropriate invocation templates and/or add different
        submission options.
        """
        # sbatch --job-name="jobname" --mem-per-cpu="MBs" --input="filename" --output="filename" --no-requeue -n "number of slots" --cpus-per-task=1 --time="minutes" script.sh
        sbatch = list(resource.sbatch)
        sbatch += ['--no-requeue']
        if self.requested_walltime:
            # SLURM uses `--time` for wall-clock time limit, expressed in minutes
            sbatch += ['--time', '%d' % self.requested_walltime.amount(minutes)]
        if self.requested_memory:
            # SLURM uses `mem_free` for memory limits; 'M' suffix allowed for Megabytes
            sbatch += ['--mem-per-cpu', '%d' % (self.requested_memory.amount(MB) / self.requested_cores) ]
        if self.stdout:
            sbatch += ['--output', '%s' % self.stdout]
        if self.stdin:
            # `self.stdin` is the full pathname on the GC3Pie client host;
            # it is copied to its basename on the execution host
            sbatch += ['--input', '%s' % os.path.basename(self.stdin)]
        if self.stderr:
            # from the sbatch(1) man page: "If both the -j y and the -e
            # options are present, Grid Engine sets but ignores the
            # error-path attribute."
            sbatch += ['-e', '%s' % self.stderr]
        if self.requested_cores != 1:
            sbatch += ['-n', ('%d' % self.requested_cores), '--cpus-per-task', '1']
        if 'jobname' in self and self.jobname:
            sbatch += ['--job-name', ('%s' % self.jobname)]
        return (sbatch, self.cmdline(resource))


    # Operation error handlers; called when transition from one state
    # to another fails.  The names are formed by suffixing the
    # corresponding `Core` method (operation) with ``_error``.

    def submit_error(self, exs):
        """
        Invocation of `Core.submit()` on this object failed;
        `exs` is a list of `Exception` objects, one for each attempted
        submission.

        If this method returns an exception object, that is raised as
        a result of the `Core.submit()`, otherwise the return value is
        ignored and `Core.submit` returns `None`.

        Default is to always return the first exception in the list
        (on the assumption that it is the root of all exceptions or
        that at least it refers to the preferred resource).  Override
        in derived classes to change this behavior.
        """
        assert len(exs) > 0, \
            "Application.submit_error called with empty list of exceptions."
        # XXX: should we choose the first or the last exception occurred?
        # vote for choosing the first, on the basis that it refers to the
        # "best" submission target
        return exs[0]


    # XXX: this method might be dangerous in that it can break the
    # `update_job_state` semantics; it's here for completeness, but we
    # should consider removing...
    def update_job_state_error(self, ex):
        """Handle exceptions that occurred during a `Core.update_job_state` call.

        If this method returns an exception object, that exception is
        processed in `Core.update_job_state()` instead of the original
        one.  Any other return value is ignored and
        `Core.update_job_state` proceeds as if no exception had
        happened.

        Argument `ex` is the exception that was raised by the backend
        during job state update.

        Default is to return `ex` unchanged; override in derived
        classes to change this behavior.

        """
        return ex


    def fetch_output_error(self, ex):
        """
        Invocation of `Core.fetch_output()` on this object failed;
        `ex` is the `Exception` that describes the error.

        If this method returns an exception object, that is raised as
        a result of the `Core.fetch_output()`, otherwise the return
        value is ignored and `Core.fetch_output` returns `None`.

        Default is to return `ex` unchanged; override in derived classes
        to change this behavior.
        """
        return ex


class _Signal(object):
    """
    Base class for representing fake signals encoding the failure
    reason for GC3Libs jobs.
    """
    def __init__(self, signum, description):
        self._signum = signum
        self.__doc__ = description
    # conversion to integer types
    def __int__(self):
        return self._signum
    def __long__(self):
        return self._signum
    # human-readable explanation
    def __str__(self):
        return "Signal %d: %s" % (self._signum, self.__doc__)

class _Signals(object):
    """
    Collection of (fake) signals used to encode termination reason in `Run.returncode`.

    ======  ============================================================
    signal  error condition
    ======  ============================================================
    125     submission to batch system failed
    124     remote error (e.g., execution node crashed, batch system misconfigured)
    123     data staging failure
    122     job killed by batch system / sysadmin
    121     job canceled by user
    ======  ============================================================

    """

    Lost = _Signal(120, "Remote site reports no information about the job")
    Cancelled = _Signal(121, "Job canceled by user")
    RemoteKill = _Signal(122, "Job killed by batch system or sysadmin")
    DataStagingFailure = _Signal(123, "Data staging failure")
    RemoteError = _Signal(124, "Unspecified remote error,"
                          " e.g., execution node crashed"
                          " or batch system misconfigured")
    SubmissionFailed = _Signal(125, "Submission to batch system failed.")

    def __contains__(self, signal):
        if (signal is not None and 120 <= int(signal) <= 125):
            return True
        else:
            return False
    def __getitem__(self, signal_num):
        if signal_num == 120:
            return Signals.Lost
        elif signal_num == 121:
            return Signals.Cancelled
        elif signal_num == 122:
            return Signals.RemoteKill
        elif signal_num == 123:
            return Signals.DataStagingFailure
        elif signal_num == 124:
            return Signals.RemoteError
        elif signal_num == 125:
            return Signals.SubmissionFailed
        else:
            raise gc3libs.exceptions.InvalidArgument("Unknown signal number %d" % signal_num)

class Run(Struct):
    """
    A specialized `dict`-like object that keeps information about
    the execution state of an `Application` instance.

    A `Run` object is guaranteed to have the following attributes:

      `log`
        A `gc3libs.utils.History` instance, recording human-readable text
        messages on events in this job's history.

      `info`
        A simplified interface for reading/writing messages to
        `Run.log`.  Reading from the `info` attribute returns the last
        message appended to `log`.  Writing into `info` appends a
        message to `log`.

      `timestamp`
        Dictionary, recording the most recent timestamp when a certain
        state was reached.  Timestamps are given as UNIX epochs.

    For properties `state`, `signal` and `returncode`, see the
    respective documentation.

    `Run` objects support attribute lookup by both the ``[...]`` and
    the ``.`` syntax; see `gc3libs.utils.Struct` for examples.
    """
    def __init__(self, initializer=None, attach=None, **keywd):
        """
        Create a new Run object; constructor accepts the same
        arguments as the `dict` constructor.

        Examples:

          1. Create a new job with default parameters::

            >>> j1 = Run()
            >>> print (j1.returncode)
            None
            >>> j1.state
            'NEW'

          2. Create a new job with additional attributes::

            >>> j2 = Run(application='GAMESS', version='2010R1')
            >>> j2.state
            'NEW'
            >>> j2.application
            'GAMESS'
            >>> j2['version']
            '2010R1'

          3. Clone an existing job object::

            >>> j3 = Run(j2)
            >>> j3.application
            'GAMESS'
            >>> j3['version']
            '2010R1'

        """
        self._ref = attach
        self._state = Run.State.NEW
        self._exitcode = None
        self._signal = None

        # to overcome the "black hole" effect
        self._execution_targets = []

        Struct.__init__(self, initializer, **keywd)

        if 'history' not in self:
            self.history = History()
        if 'timestamp' not in self:
            self.timestamp = OrderedDict()


    @defproperty
    def info():
        """
        A simplified interface for reading/writing entries into `history`.

        Setting the `info` attribute appends a message to the log::

           >>> j1 = Run()
           >>> j1.info = 'a message'
           >>> j1.info = 'a second message'

        Getting the value of the `info` attribute returns the last
        message entered in the log::

          >>> j1.info # doctest: +ELLIPSIS
          u'a second message ...'

        """
        def fget(self):
            return self.history.last()
        def fset(self, value):
            self.history.append(unicode(value))
        return locals()

    # states that a `Run` can be in
    State = Enum(
        'NEW',       # Job has not yet been submitted/started
        'SUBMITTED', # Job has been sent to execution resource
        'STOPPED',   # trap state: job needs manual intervention
        'RUNNING',   # job is executing on remote resource
        'TERMINATING', # remote job execution finished, output not yet retrieved
        'TERMINATED',# job execution finished (correctly or not) and will not be resumed
        'UNKNOWN',   # job info not found or lost track of job (e.g., network error or invalid job ID)
        )

    class Arch(object):
        """
        Processor architectures, for use as values in the
        `requested_architecture` field of the `Application` class
        constructor.

        The following values are currently defined:

        `X86_64`
          64-bit Intel/AMD/VIA x86 processors in 64-bit mode.

        `X86_32`
          32-bit Intel/AMD/VIA x86 processors in 32-bit mode.
        """
        X86_64 = "x86_64"
        X86_32 = "i686"

        # the following method makes this class read-only,
        # thus preventing users accidentally overwriting the
        # value of constants above...
        def __setattr__(self, name, value):
            raise TypeError("Cannot overwrite value of constant '%s'" % name)

    @defproperty
    def state():
        """
        The state a `Run` is in.

        The value of `Run.state` must always be a value from the
        `Run.State` enumeration, i.e., one of the following values.

        +---------------+--------------------------------------------------------------+----------------------+
        |Run.State value|purpose                                                       |can change to         |
        +===============+==============================================================+======================+
        |NEW            |Job has not yet been submitted/started (i.e., gsub not called)|SUBMITTED (by gsub)   |
        +---------------+--------------------------------------------------------------+----------------------+
        |SUBMITTED      |Job has been sent to execution resource                       |RUNNING, STOPPED      |
        +---------------+--------------------------------------------------------------+----------------------+
        |STOPPED        |Trap state: job needs manual intervention (either user-       |TERMINATING(by gkill),|
        |               |or sysadmin-level) to resume normal execution                 |SUBMITTED (by miracle)|
        +---------------+--------------------------------------------------------------+----------------------+
        |RUNNING        |Job is executing on remote resource                           |TERMINATING           |
        +---------------+--------------------------------------------------------------+----------------------+
        |TERMINATING    |Job has finished execution on remote resource;                |TERMINATED            |
        |               |output not yet retrieved                                      |                      |
        +---------------+--------------------------------------------------------------+----------------------+
        |TERMINATED     |Job execution is finished (correctly or not)                  |None: final state     |
        |               |and will not be resumed; output has been retrieved            |                      |
        +---------------+--------------------------------------------------------------+----------------------+

        When a :class:`Run` object is first created, it is assigned
        the state NEW.  After a successful invocation of
        `Core.submit()`, it is transitioned to state SUBMITTED.
        Further transitions to RUNNING or STOPPED or TERMINATED state,
        happen completely independently of the creator progra; the
        `Core.update_job_state()` call provides updates on the status
        of a job.

        The STOPPED state is a kind of generic "run time error" state:
        a job can get into the STOPPED state if its execution is
        stopped (e.g., a SIGSTOP is sent to the remote process) or
        delayed indefinitely (e.g., the remote batch system puts the
        job "on hold"). There is no way a job can get out of the
        STOPPED state automatically: all transitions from the STOPPED
        state require manual intervention, either by the submitting
        user (e.g., cancel the job), or by the remote systems
        administrator (e.g., by releasing the hold).

        The TERMINATED state is the final state of a job: once a job
        reaches it, it cannot get back to any other state. Jobs reach
        TERMINATED state regardless of their exit code, or even if a
        system failure occurred during remote execution; actually,
        jobs can reach the TERMINATED status even if they didn't run
        at all, for example, in case of a fatal failure during the
        submission step.
        """
        def fget(self):
            return self._state
        def fset(self, value):
            assert value in Run.State, \
                ("Value '%s' is not a legal `gc3libs.Run.State` value." % value)
            if self._state != value:
                self.state_last_changed = time.time()
                self.timestamp[value] = time.time()
                self.history.append(value)
                if self._ref is not None:
                    # mark as changed
                    self._ref.changed = True
                    # invoke state-transition method
                    handler = value.lower()
                    gc3libs.log.debug(
                        "Calling state-transition handler '%s' on %s ..."
                        % (handler, self._ref))
                    getattr(self._ref, handler)()
            self._state = value
        return locals()


    def in_state(self, *names):
        """
        Return `True` if the `Run` state matches any of the given names.

        In addition to the states from `Run.State`:class:, the two
        additional names ``ok`` and ``failed`` are also accepted, with
        the following meaning:

        * ``ok``: state is `TERMINATED` and `returncode` is 0.

        * ``failed``: state is `TERMINATED` and `returncode` is non-zero.
        """
        state = self.state
        if state in names:
            return True
        elif ('ok' in names
              and state == Run.State.TERMINATED and self.returncode == 0):
            return True
        elif ('failed' in names
              and state == Run.State.TERMINATED and self.returncode != 0):
            return True
        else:
            return False


    @defproperty
    def signal():
        """
        The "signal number" part of a `Run.returncode`, see
        `os.WTERMSIG` for details.

        The "signal number" is a 7-bit integer value in the range
        0..127; value `0` is used to mean that no signal has been
        received during the application runtime (i.e., the application
        terminated by calling ``exit()``).

        The value represents either a real UNIX system signal, or a
        "fake" one that GC3Libs uses to represent Grid middleware
        errors (see `Run.Signals`).
        """
        def fget(self):
            return self._signal
        def fset(self, value):
            if value is None:
                self._signal = None
            else:
                self._signal = int(value) & 0x7f
        return (locals())


    @defproperty
    def exitcode():
        """
        The "exit code" part of a `Run.returncode`, see `os.WEXITSTATUS`.
        This is an 8-bit integer, whose meaning is entirely
        application-specific.  (However, the value `255` is often used
        to mean that an error has occurred and the application could
        not end its execution normally.)
        """
        def fget(self):
            return self._exitcode
        def fset(self, value):
            if value is None:
                self._exitcode = None
            else:
                self._exitcode = int(value) & 0xff
        return (locals())


    @defproperty
    def returncode():
        """
        The `returncode` attribute of this job object encodes the
        `Run` termination status in a manner compatible with the POSIX
        termination status as implemented by `os.WIFSIGNALED` and
        `os.WIFEXITED`.

        However, in contrast with POSIX usage, the `exitcode` and the
        `signal` part can *both* be significant: in case a Grid
        middleware error happened *after* the application has
        successfully completed its execution.  In other words,
        `os.WEXITSTATUS(returncode)` is meaningful iff
        `os.WTERMSIG(returncode)` is 0 or one of the pseudo-signals
        listed in `Run.Signals`.

        `Run.exitcode` and `Run.signal` are combined to form the
        return code 16-bit integer as follows (the convention appears
        to be obeyed on every known system):

           +------+------------------------------------+
           |Bit   |Encodes...                          |
           +======+====================================+
           |0..7  |signal number                       |
           +------+------------------------------------+
           |8     |1 if program dumped core.           |
           +------+------------------------------------+
           |9..16 |exit code                           |
           +------+------------------------------------+

        *Note:* the "core dump bit" is always 0 here.

        Setting the `returncode` property sets `exitcode` and
        `signal`; you can either assign a `(signal, exitcode)` pair to
        `returncode`, or set `returncode` to an integer from which the
        correct `exitcode` and `signal` attribute values are
        extracted::

           >>> j = Run()
           >>> j.returncode = (42, 56)
           >>> j.signal
           42
           >>> j.exitcode
           56

           >>> j.returncode = 137
           >>> j.signal
           9
           >>> j.exitcode
           0

        See also `Run.exitcode` and `Run.signal`.
        """
        def fget(self):
            if self.exitcode is None and self.signal is None:
                return None
            if self.exitcode is None:
                exitcode = -1
            else:
                exitcode = self.exitcode
            if self.signal is None:
                signal = 0
            else:
                signal = self.signal
            return (exitcode << 8) | signal
        def fset(self, value):
            if value is None:
                self.signal = None
                self.exitcode = None
            else:
                try:
                    # `value` can be a tuple `(signal, exitcode)`
                    self.signal = int(value[0])
                    self.exitcode = int(value[1])
                except (TypeError, ValueError):
                    self.exitcode = (int(value) >> 8) & 0xff
                    self.signal = int(value) & 0x7f
                # ensure values are within allowed range
                self.exitcode &= 0xff
                self.signal &= 0x7f
        return (locals())

    # `Run.Signals` is an instance of global class `_Signals`
    Signals = _Signals()


    @staticmethod
    def shellexit_to_returncode(rc):
        """
        Convert a shell exit code to a POSIX process return code.

        A POSIX shell represents the return code of the last-run
        program within its exit code as follows:

        * If the program was terminated by signal `N`, the shell exits
          with code 128+N,

        * otherwise, if the program terminated with exit code C, the
          shell exits with code C.
        """
        # only the less significant 8 bits matter
        rc &= 0xff
        if rc > 128:
            # terminated by signal N is encoded as 128+N
            return (rc-128, -1)
        else:
            # regular exit
            return (0, rc)

# Factory method to create an Engine instance

def create_engine(*conf_files, **extra_args):
    """
    Create and return a `gc3libs.core.Engine`:class: class.

    It accepts an optional list of configuration filenames. If the
    filenames contain a `~` or a variable name, it will be expanded
    automatically.

    Called without arguments, a configuration file will be searched in
    ~/.gc3/gc3pie.conf (and used if found).

    :param conf_files:    List of configuration files to read.
    :param store:         See the like-named argument to the `Engine`:class: constructor.
    :param can_submit:    See the like-named argument to the `Engine`:class: constructor.
    :param can_retrieve:  See the like-named argument to the `Engine`:class: constructor.
    :param max_in_flight: See the like-named argument to the `Engine`:class: constructor.
    :param max_submitted: See the like-named argument to the `Engine`:class: constructor.

    Any extra keyword argument is passed unchanged to the
    `Configuration`:class: constructor.

    """
    store = extra_args.pop('store', None)
    can_submit = extra_args.pop('can_submit', True)
    can_retrieve = extra_args.pop('can_retrieve', True)
    max_in_flight = extra_args.pop('max_in_flight', 0)
    max_submitted = extra_args.pop('max_submitted', 0)

    from gc3libs.config import Configuration
    from gc3libs.core import Core, Engine
    conf_files = [
        os.path.expandvars(
            os.path.expanduser(fname)) for fname in conf_files]
    if not conf_files:
        conf_files = Default.CONFIG_FILE_LOCATIONS[:]

    if not 'auto_enable_auth' in extra_args:
        extra_args['auto_enable_auth'] = True

    cfg = Configuration(*conf_files, **extra_args)
    core = Core(cfg)
    engine = Engine(core, store=store,
                    can_submit=can_submit, can_retrieve=can_retrieve,
                    max_in_flight=max_in_flight, max_submitted=max_submitted)

    return engine


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)
